# -*- coding: utf-8 -*-
"""
Copyright © 2024 Wacom Authors. All Rights Reserved.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

   http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""
import json
import math
import uuid
from pathlib import Path
from statistics import StatisticsError
from typing import Dict, Any, Tuple, List

import numpy as np
import pytest

from uim.codec.parser.base import SupportedFormats
from uim.codec.parser.uim import UIMParser
from uim.model import UUIDIdentifier
from uim.model.helpers.policy import HandleMissingDataPolicy
from uim.model.helpers.schema_content_extractor import uim_schema_semantics_from
from uim.model.helpers.text_extractor import uim_extract_text_and_semantics_from, uim_extract_text_and_semantics
from uim.model.ink import InkModel, InkTree, ViewTree
from uim.model.inkdata.brush import VectorBrush, BrushPolygonUri
from uim.model.inkdata.strokes import Spline, LayoutMask, Style, Stroke, InkStrokeAttributeType
from uim.model.inkinput.inputdata import InkSensorType, InputDevice, Environment, InkInputProvider, InkInputType, \
    SensorChannel, InkSensorMetricType, SensorChannelsContext, SensorContext, InputContext
from uim.model.inkinput.sensordata import SensorData, InkState, ChannelData
from uim.model.semantics import schema
from uim.model.semantics.node import StrokeNode, StrokeGroupNode
from uim.model.semantics.schema import (CommonViews, SemanticTriple, DEVICE_MANUFACTURER_PROPERTY,
                                        DEVICE_MODEL_PROPERTY, DEVICE_SERIAL_NUMBER_PROPERTY, DOCUMENT_AUTHOR_OBJECT,
                                        DOCUMENT_LOCALE_OBJECT)
from uim.utils.analyser import as_strided_array
from uim.utils.matrix import Matrix4x4
from uim.utils.statistics import StatisticsAnalyzer
from uim.utils.stroke_resampling import CurvatureBasedInterpolationCalculator, SplineHandler, \
    StrokeResamplerInkModelWrapper

# Test data directory
ink_dir: Path = Path(__file__).parent / '../ink'
uim300_data_dir: Path = ink_dir / 'uim_3.0.0'
uim310_data_dir: Path = ink_dir / 'uim_3.1.0'
schema_data_dir: Path = ink_dir / 'schemas'
reference_text_dir: Path = ink_dir / 'reference_text.uim_3.0.0_json'
ENVIRONMENT: str = "qa-environment"
LANGUAGE: str = "en_US"
AUTHOR: str = "John Doe"


def uim_300_files() -> list:
    return [f for f in uim300_data_dir.iterdir() if f.is_file() and f.name.endswith('.uim')]


def uim_310_files() -> list:
    return [f for f in uim310_data_dir.iterdir() if f.is_file() and f.name.endswith('.uim')]




def simulate_pressure(num_points: int, max_pressure: float = 1.0, noise_level: float = 0.1):
    x = np.linspace(0, np.pi, num_points)
    pressure = (np.sin(x) + 1) / 2 * max_pressure

    # Add a small constant to start at 0.01
    pressure = np.clip(pressure, 0.01, None)

    # Add Gaussian noise to simulate variation
    pressure += np.random.normal(0, noise_level, num_points)

    # Ensure no negative values and start from 0.01
    pressure = np.clip(pressure, 0.01, None)
    return pressure


def create_random_sensor_data(x_0: float = 100., y_0: float = 100., pressure_0: float = 0.3, noise_level: float = 1,
                              num_points: int = 5, sample_rate: int = 120) \
        -> Tuple[List[float], List[float], List[float], List[float]]:
    timestamps = np.linspace(0., num_points * (1. / sample_rate), num_points)
    # Add Gaussian noise to simulate handwriting variation
    x_values: List[float] = []
    y_values: List[float] = []
    last_x: float = x_0
    last_y: float = y_0
    for i in range(num_points):
        x_val = last_x + abs(np.random.normal(0, noise_level))
        y_val = last_y + np.random.normal(0, noise_level)
        x_values.append(x_val)
        y_values.append(y_val)
        last_x = x_val
        last_y = y_val
    pressure = simulate_pressure(num_points, max_pressure=pressure_0, noise_level=0.1)
    return timestamps.tolist(), x_values, y_values, pressure.tolist()


def created_model(num_point: int = 100, number_strokes: int = 1, with_timestamps: bool = True,
                  with_pressure: bool = True, add_sensor_data: bool = True) -> InkModel:
    # Create the model
    ink_model: InkModel = InkModel(version=SupportedFormats.UIM_VERSION_3_1_0.value)
    # Setting a unit scale factor
    ink_model.unit_scale_factor = 1.
    # Using a 4x4 matrix for scaling
    ink_model.transform = Matrix4x4.create_scale(1.)

    # Properties are added as key-value pairs
    ink_model.properties.append((DOCUMENT_AUTHOR_OBJECT, AUTHOR))
    ink_model.properties.append((DOCUMENT_LOCALE_OBJECT, LANGUAGE))

    # Create an environment
    env: Environment = Environment()
    env.properties.append(("env.name", ENVIRONMENT))
    ink_model.input_configuration.environments.append(env)

    # Ink input provider pen
    provider: InkInputProvider = InkInputProvider(input_type=InkInputType.PEN)
    ink_model.input_configuration.ink_input_providers.append(provider)

    # Input device is the sensor (pen tablet, screen, etc.)
    input_device: InputDevice = InputDevice()
    input_device.properties.append((DEVICE_SERIAL_NUMBER_PROPERTY, uuid.uuid4().hex))
    input_device.properties.append((DEVICE_MANUFACTURER_PROPERTY, "Wacom"))
    input_device.properties.append((DEVICE_MODEL_PROPERTY, "Artificial Data"))

    ink_model.input_configuration.devices.append(input_device)

    # Create a group of sensor channels
    sensor_channels: list = [
        SensorChannel(channel_type=InkSensorType.X, metric=InkSensorMetricType.LENGTH, resolution=1.0),
        SensorChannel(channel_type=InkSensorType.Y, metric=InkSensorMetricType.LENGTH, resolution=1.0)
    ]
    time_idx: int = 0
    pressure_idx: int = 0
    idx: int = 2
    if with_pressure:
        sensor_channels.append(SensorChannel(channel_type=InkSensorType.PRESSURE, metric=InkSensorMetricType.FORCE,
                                             resolution=1.0))
        pressure_idx = idx
        idx += 1
    if with_timestamps:
        sensor_channels.append(SensorChannel(channel_type=InkSensorType.TIMESTAMP, metric=InkSensorMetricType.TIME,
                                             resolution=1.0))
        time_idx = idx
    scc: SensorChannelsContext = SensorChannelsContext(channels=sensor_channels,
                                                       ink_input_provider_id=provider.id,
                                                       input_device_id=input_device.id)

    # Add all sensor channel contexts
    sensor_context: SensorContext = SensorContext()
    sensor_context.add_sensor_channels_context(scc)
    ink_model.input_configuration.sensor_contexts.append(sensor_context)

    # Create the input context using the Environment and the Sensor Context
    input_context: InputContext = InputContext(environment_id=env.id, sensor_context_id=sensor_context.id)
    ink_model.input_configuration.input_contexts.append(input_context)

    # We need to define a brush polygon
    # Add a brush specified with shape Uris
    poly_uris: list = [
        BrushPolygonUri("will://brush/3.0/shape/Circle?precision=20&radius=1", 0.),
        BrushPolygonUri("will://brush/3.0/shape/Ellipse?precision=20&radiusX=1&radiusY=0.5", 4.0)
    ]

    # Create a vector brush
    vector_brush_0: VectorBrush = VectorBrush("app://qa-test-app/vector-brush/Brush", poly_uris)

    # Add it to the model
    ink_model.brushes.add_vector_brush(vector_brush_0)

    # Create some style
    style: Style = Style(brush_uri=vector_brush_0.name)
    style.path_point_properties.red = 0.
    style.path_point_properties.green = 0.0
    style.path_point_properties.blue = 1.0
    style.path_point_properties.alpha = 1.0
    # First you need a root group to contain the strokes
    root: StrokeGroupNode = StrokeGroupNode(UUIDIdentifier.id_generator())
    # Create a view tree for the HWR
    hwr_tree: ViewTree = ViewTree(CommonViews.HWR_VIEW.value)
    # Create an ink tree
    ink_model.ink_tree = InkTree()
    # Assign the group as the root of the main ink tree
    ink_model.ink_tree.root = root
    hwr_root: StrokeGroupNode = StrokeGroupNode(UUIDIdentifier.id_generator())
    hwr_tree.root = hwr_root
    # The hwr root denotes a word
    hwr_text_line: StrokeGroupNode = StrokeGroupNode(UUIDIdentifier.id_generator())
    hwr_text_region: StrokeGroupNode = StrokeGroupNode(UUIDIdentifier.id_generator())
    hwr_root.add(hwr_text_region)
    hwr_text_region.add(hwr_text_line)

    x_offset: float = 0.
    y_offset: float = 0.
    for _ in range(number_strokes):
        hwr_word: StrokeGroupNode = StrokeGroupNode(UUIDIdentifier.id_generator())
        # Generate a unique identifier for the sensor data
        sensor_data_id: uuid.UUID = UUIDIdentifier.id_generator()
        # Create sensor data
        sensor_data_i: SensorData = SensorData(sensor_data_id, input_context_id=input_context.id,
                                               state=InkState.PLANE)
        # Generate some random sensor data
        timestamps, x_new, y_new, pressure = create_random_sensor_data(x_0=x_offset, y_0=y_offset, num_points=num_point)
        sensor_data_i.add_data(sensor_channels[0], x_new)
        sensor_data_i.add_data(sensor_channels[1], y_new)
        if with_pressure:
            sensor_data_i.add_data(sensor_channels[pressure_idx], pressure)
        if with_timestamps:
            sensor_data_i.add_timestamp_data(sensor_channels[time_idx], timestamps)
        if add_sensor_data:
            # Add sensor data to the model
            ink_model.sensor_data.add(sensor_data_i)

        # Specify the layout of the stroke data, in this case the stroke will have variable X, Y and Size properties.
        layout_mask: int = LayoutMask.X.value | LayoutMask.Y.value | LayoutMask.SIZE.value

        # Provide the stroke data - in this case 4 data points, each consisting of X, Y, Size
        path: list = [x_new[0], y_new[0], 1.0]
        for x, y, p in zip(x_new, y_new, pressure):
            path.append(x)
            path.append(y)
            path.append(1. + p)
        path.append(x_new[-1])
        path.append(y_new[-1])
        path.append(1.0)
        # Create a spline object from the path data
        spline: Spline = Spline(layout_mask, path)
        # Create a stroke object. Note that it just exists, but is not in the model yet.
        stroke_i: Stroke = Stroke(sid=UUIDIdentifier.id_generator(), spline=spline, style=style,
                                  sensor_data_id=sensor_data_id if add_sensor_data else None)
        # Add a node for stroke
        root.add(StrokeNode(stroke_i))
        hwr_word.add(StrokeNode(stroke_i))
        hwr_text_line.add(hwr_word)
        ink_model.knowledge_graph.append(SemanticTriple(hwr_word.uri, schema.IS, schema.SegmentationSchema.WORD))
        ink_model.knowledge_graph.append(SemanticTriple(hwr_word.uri, schema.SegmentationSchema.HAS_CONTENT, '-'))
        x_offset += 100.
        y_offset += 100.
    ink_model.add_view(hwr_tree)
    ink_model.knowledge_graph.append(SemanticTriple(hwr_text_region.uri, schema.IS,
                                                    schema.SegmentationSchema.TEXT_REGION))
    ink_model.knowledge_graph.append(SemanticTriple(hwr_text_line.uri, schema.IS, schema.SegmentationSchema.TEXT_LINE))
    ink_model.calculate_bounds_recursively(ink_model.ink_tree.root)
    ink_model.calculate_bounds_recursively(hwr_tree.root)
    return ink_model


@pytest.mark.parametrize('path', uim_310_files())
def test_uim_3_1_0_semantics(path: Path):
    parser: UIMParser = UIMParser()
    ink_model: InkModel = parser.parse(path)
    if reference_text_dir.exists():
        with open(reference_text_dir, 'r') as f:
            reference_text: Dict[str, str] = json.loads(f.read())
    else:
        reference_text: Dict[str, str] = {}
    if ink_model.has_knowledge_graph() and ink_model.has_tree(CommonViews.HWR_VIEW.value):
        words, entities, text = uim_extract_text_and_semantics_from(ink_model, CommonViews.HWR_VIEW.value)
        assert len(words) > 0
        assert len(entities) > 0
        assert len(text) > 0
        if path.name in reference_text:
            assert text == reference_text[path.name]


def test_uim_3_1_0_math_structures():
    parser: UIMParser = UIMParser()
    ink_model: InkModel = parser.parse(schema_data_dir / 'math-structures.uim')
    if ink_model.has_knowledge_graph() and ink_model.has_tree(CommonViews.HWR_VIEW.value):
        elements = uim_schema_semantics_from(ink_model, CommonViews.HWR_VIEW.value)
        for element in elements:
            assert isinstance(element, Dict)
            assert 'type' in element
            assert element['type'] in [schema.SegmentationSchema.ROOT, schema.MathStructureSchema.MATH_BLOCK_STRUCTURES,
                                       schema.MathStructureSchema.SYMBOL,
                                       schema.MathStructureSchema.NUMBER,
                                       schema.MathStructureSchema.DIGIT,
                                       schema.MathStructureSchema.OPERATION,
                                       schema.MathStructureSchema.GROUP,
                                       schema.MathStructureSchema.FRACTION,
                                       schema.MathStructureSchema.RELATION,
                                       schema.MathStructureSchema.FENCE,
                                       schema.MathStructureSchema.MATHEMATICAL_TERM,
                                       schema.MathStructureSchema.SEPARATOR,
                                       schema.MathStructureSchema.SUPERSCRIPT,
                                       schema.MathStructureSchema.RELATION_SYMBOL,
                                       schema.MathStructureSchema.OPERATOR_SYMBOL]
            for path_id in element['path_id']:
                assert isinstance(path_id, uuid.UUID)
            assert 'bounding_box' in element
            assert isinstance(element['bounding_box'], Dict)
            assert isinstance(element['bounding_box']['x'], float)
            assert isinstance(element['bounding_box']['y'], float)
            assert isinstance(element['bounding_box']['width'], float)
            assert isinstance(element['bounding_box']['height'], float)
            assert 'attributes' in element
            for attribute in element['attributes']:
                assert isinstance(attribute[0], str)
                assert isinstance(attribute[1], (str, int, float))
  

@pytest.mark.parametrize('path', uim_310_files())
def test_uim_3_1_0_semantics(path: Path):
    parser: UIMParser = UIMParser()
    ink_model: InkModel = parser.parse(path)
    if reference_text_dir.exists():
        with open(reference_text_dir, 'r') as f:
            reference_text: Dict[str, str] = json.loads(f.read())
    else:
        reference_text: Dict[str, str] = {}
    if ink_model.has_knowledge_graph() and ink_model.has_tree(CommonViews.HWR_VIEW.value):
        words, entities, text = uim_extract_text_and_semantics_from(ink_model, CommonViews.HWR_VIEW.value)
        if words is None or len(words) == 0:
            print()
        assert len(words) > 0
        assert len(entities) > 0
        assert len(text) > 0
        if path.name in reference_text:
            assert text == reference_text[path.name]
        with path.open('rb') as f:
            uim_bytes: bytes = f.read()
            words_2, entities_2, text_2 = uim_extract_text_and_semantics(uim_bytes, CommonViews.HWR_VIEW.value)
            assert text == text_2
            assert len(words) == len(words_2)
            for w1, w2 in zip(words, words_2):
                assert w1['text'] == w2['text']
                assert w1['path_id'] == w2['path_id']
                assert w1['bounding_box'] == w2['bounding_box']
            assert len(entities) == len(entities_2)
            assert entities == entities_2


@pytest.mark.parametrize('path', uim_300_files())
def test_uim_semantics_schema(path: Path):
    """
    Test the semantics schema extraction.
    """
    parser: UIMParser = UIMParser()
    ink_model: InkModel = parser.parse(path)
    if ink_model.has_knowledge_graph() and ink_model.has_tree(CommonViews.HWR_VIEW.value):
        elements = uim_schema_semantics_from(ink_model, CommonViews.HWR_VIEW.value)
        assert len(elements) > 0
        for element in elements:
            assert isinstance(element, Dict)
            assert 'type' in element
            for path_id in element['path_id']:
                assert isinstance(path_id, uuid.UUID)
            assert 'bounding_box' in element
            assert isinstance(element['bounding_box'], Dict)
            assert isinstance(element['bounding_box']['x'], float)
            assert isinstance(element['bounding_box']['y'], float)
            assert isinstance(element['bounding_box']['width'], float)
            assert isinstance(element['bounding_box']['height'], float)
            assert 'attributes' in element
            for attribute in element['attributes']:
                assert isinstance(attribute[0], str)
                assert isinstance(attribute[1], (str, int, float))


@pytest.mark.parametrize('path', uim_310_files())
def test_uim_3_1_0_interpolation(path: Path, point_threshold: int = 15):
    """
    Test the strided array method.
    """
    parser: UIMParser = UIMParser()
    ink_model: InkModel = parser.parse(path)
    strokes_as_strided_array, strided_strokes_layout = ink_model.get_strokes_as_strided_array_extended()
    calculator: CurvatureBasedInterpolationCalculator = CurvatureBasedInterpolationCalculator()
    for si, stroke in enumerate(strokes_as_strided_array):
        resampled_stroke = SplineHandler.process(stroke, strided_strokes_layout, point_threshold, calculator)
        calculator.reset_state()
        assert len(resampled_stroke) > 0
        
        
def test_interpolation():
    """
    Test the strided array method.
    """
    calculator: CurvatureBasedInterpolationCalculator = CurvatureBasedInterpolationCalculator()
    layout: List[InkStrokeAttributeType] = [
        InkStrokeAttributeType.SPLINE_X, InkStrokeAttributeType.SPLINE_Y, InkStrokeAttributeType.SENSOR_TIMESTAMP,
        InkStrokeAttributeType.SENSOR_PRESSURE
    ]
    ink_model = created_model(num_point=100, number_strokes=2)
    process_layout: List[InkStrokeAttributeType] = [l for l in InkStrokeAttributeType]
    for stroke in ink_model.strokes:
        strokes_as_strided_array = stroke.as_strided_array_extended(ink_model)
        resampled_stroke = SplineHandler.process(strokes_as_strided_array, process_layout, 15, calculator)
        assert len(resampled_stroke) > 0


@pytest.mark.parametrize('path', uim_310_files())
def test_uim_3_1_0_resample(path: Path, point_threshold: int = 20):
    stroke_resampler = StrokeResamplerInkModelWrapper()
    parser: UIMParser = UIMParser()
    ink_model: InkModel = parser.parse(path)
    stroke_resampler.resample_ink_model(ink_model, point_threshold)


def test_artificial_data():
    """
    Test the fix_spline method.
    """
    ink_model_art: InkModel = created_model()
    strokes_as_strided_array, strided_strokes_layout = ink_model_art.get_strokes_as_strided_array_extended()
    calculator: CurvatureBasedInterpolationCalculator = CurvatureBasedInterpolationCalculator()
    for point_threshold in range(5, 20):
        for si, stroke in enumerate(strokes_as_strided_array):
            resampled_stroke = SplineHandler.process(stroke, strided_strokes_layout, point_threshold, calculator)
            calculator.reset_state()
            assert len(resampled_stroke) > 0


def test_strided_array():
    """
    Test the strided array method.
    """
    ink_model: InkModel = created_model(with_timestamps=False)
    strokes_as_strided_array, _ = ink_model.get_strokes_as_strided_array(policy=HandleMissingDataPolicy.SKIP_STROKE)
    assert len(strokes_as_strided_array) == 0
    # Check exception
    with pytest.raises(ValueError):
        ink_model.get_strokes_as_strided_array(policy=HandleMissingDataPolicy.THROW_EXCEPTION)
    ink_model: InkModel = created_model(with_pressure=False)
    strokes_as_strided_array, _ = ink_model.get_strokes_as_strided_array(policy=HandleMissingDataPolicy.SKIP_STROKE)
    assert len(strokes_as_strided_array) == 0
    # Check all exceptions
    with pytest.raises(ValueError):
        ink_model.get_strokes_as_strided_array(policy=HandleMissingDataPolicy.THROW_EXCEPTION)
    with pytest.raises(ValueError):
        ink_model.get_strokes_as_strided_array(policy=HandleMissingDataPolicy.FILL_WITH_NAN)
    with pytest.raises(ValueError):
        ink_model.get_strokes_as_strided_array([InkStrokeAttributeType.SPLINE_X, InkStrokeAttributeType.SPLINE_Y])
    ink_model: InkModel = created_model()
    strokes_as_strided_array, strided_strokes_layout = ink_model.get_strokes_as_strided_array()
    assert len(strokes_as_strided_array) > 0
    assert len(strided_strokes_layout) > 0


def test_strided_array_extended():
    """
    Test the strided array method.
    """
    num_points: int = 100
    ink_model: InkModel = created_model(num_point=num_points)
    # Spline x, y are mandatory
    with pytest.raises(ValueError):
        ink_model.get_strokes_as_strided_array_extended([InkStrokeAttributeType.SENSOR_TIMESTAMP,
                                                         InkStrokeAttributeType.SENSOR_PRESSURE])
    strokes_as_strided_array, strided_strokes_layout = ink_model.get_strokes_as_strided_array_extended()
    for array in strokes_as_strided_array:
        assert len(array) == num_points * len(strided_strokes_layout)
    assert len(strided_strokes_layout) > 0
    ink_model: InkModel = created_model(with_timestamps=False)
    strokes_as_strided_array, _ = ink_model.get_strokes_as_strided_array_extended(
        policy=HandleMissingDataPolicy.SKIP_STROKE
    )
    assert len(strokes_as_strided_array) == 0
    # Check exception
    with pytest.raises(ValueError):
        ink_model.get_strokes_as_strided_array(policy=HandleMissingDataPolicy.THROW_EXCEPTION)
    ink_model: InkModel = created_model(with_pressure=False)
    strokes_as_strided_array, _ = ink_model.get_strokes_as_strided_array(policy=HandleMissingDataPolicy.SKIP_STROKE)
    assert len(strokes_as_strided_array) == 0
    # Check exception
    with pytest.raises(ValueError):
        ink_model.get_strokes_as_strided_array(policy=HandleMissingDataPolicy.THROW_EXCEPTION)


def test_sensor_lookup():
    """
    Test the sensor lookup method.
    """
    num_points: int = 100
    ink_model: InkModel = created_model(num_point=num_points)
    for stroke in ink_model.strokes:
        res = ink_model.sensor_data_lookup(stroke=stroke, ink_sensor_type=InkSensorType.TIMESTAMP)
        assert len(res) == num_points
        res = ink_model.sensor_data_lookup(stroke=stroke, ink_sensor_type=InkSensorType.PRESSURE)
        assert len(res) == num_points
    ink_model: InkModel = created_model(with_timestamps=False)
    for stroke in ink_model.strokes:
        res = ink_model.sensor_data_lookup(stroke=stroke, ink_sensor_type=InkSensorType.TIMESTAMP)
        assert len(res) == 0
        inst = ink_model.sensor_data_lookup(stroke=stroke, ink_sensor_type=InkSensorType.TIMESTAMP,
                                            return_channel_data_instance=True)
        assert inst is None
        res = ink_model.sensor_data_lookup(stroke=stroke, ink_sensor_type=InkSensorType.PRESSURE)
        assert len(res) == num_points
        inst = ink_model.sensor_data_lookup(stroke=stroke, ink_sensor_type=InkSensorType.PRESSURE,
                                            return_channel_data_instance=True)
        assert isinstance(inst, ChannelData)


def test_two_strokes():
    """
    Test the two strokes method.
    """
    model_analyser: StatisticsAnalyzer = StatisticsAnalyzer()
    artificial_model: InkModel = created_model(number_strokes=2)
    try:
        stats: Dict[str, Any] = model_analyser.analyze(artificial_model)

        assert len(stats) > 0
        assert len(stats['properties']) == 2
        # Check the properties
        assert stats['properties'][DOCUMENT_AUTHOR_OBJECT]['documents_count'] == 0
        assert stats['properties'][DOCUMENT_AUTHOR_OBJECT]['values'][AUTHOR]['count'] == 1
        assert stats['properties'][DOCUMENT_AUTHOR_OBJECT]['values'][AUTHOR]['percent'] == 0.0
        assert stats['properties'][DOCUMENT_LOCALE_OBJECT]['documents_count'] == 0
        assert stats['properties'][DOCUMENT_LOCALE_OBJECT]['values'][LANGUAGE]['count'] == 1
        assert stats['properties'][DOCUMENT_LOCALE_OBJECT]['values'][LANGUAGE]['percent'] == 0.0
        first_key = list(stats['input_devices'].keys())[0]
        assert stats['input_devices'][first_key]['strokes_count'] == 2
        assert stats['input_devices'][first_key]['percent'] == 100.0
        assert len(stats['sensor_channels']) == 4
        assert len(stats['document_bounds']) == 6
        # Check the sensor channels
        assert stats['sensor_channels']['X']['min'] > 0.
        assert stats['sensor_channels']['PRESSURE']['min'] > 0.
        assert stats['sensor_channels']['TIMESTAMP']['min'] == 0.0
        assert stats['sensor_channels']['X']['max'] > 0.
        assert stats['sensor_channels']['Y']['max'] > 0.
        assert stats['sensor_channels']['PRESSURE']['max'] > 0.
        assert stats['sensor_channels']['TIMESTAMP']['max'] > 0.
        assert stats['sensor_channels']['X']['mean'] > 0.
        assert stats['sensor_channels']['Y']['mean'] > 0.
        assert stats['sensor_channels']['PRESSURE']['mean'] > 0.
        assert stats['sensor_channels']['TIMESTAMP']['mean'] > 0.
        assert stats['sensor_channels']['X']['median'] > 0.
        assert stats['sensor_channels']['Y']['median'] > 0.
        assert stats['sensor_channels']['PRESSURE']['median'] > 0.
        assert stats['sensor_channels']['TIMESTAMP']['median'] > 0.
        # Check the strokes
        assert stats['strokes_count'] == 2
        assert stats['points_count']['total'] == 204
        assert stats['points_count']['min'] == 102
        assert stats['points_count']['max'] == 102
        assert stats['points_count']['mean'] == 102
        assert stats['points_count']['std'] == 0.0
        assert stats['points_count']['median'] == 102.0
        assert stats['document_stats']['min_area'] > 0.
        assert stats['document_stats']['max_area'] > 0.
        assert stats['sampling_rate'] == 0.01
        # Check the brush, env, input devices and input providers
        first_key = list(stats['brushes'].keys())[0]
        assert stats['brushes'][first_key]['strokes_count'] == 2
        assert stats['brushes'][first_key]['percent'] == 100.0
        first_key = list(stats['envs'].keys())[0]
        assert stats['envs'][first_key]['strokes_count'] == 2
        assert stats['envs'][first_key]['percent'] == 100.0
        assert len(stats['input_devices']) == 1
        first_key = list(stats['input_devices'].keys())[0]
        assert stats['input_devices'][first_key]['strokes_count'] == 2
        assert stats['input_devices'][first_key]['percent'] == 100.0
        assert len(stats['input_providers']) == 1
        first_key = list(stats['input_providers'].keys())[0]
        assert stats['input_providers'][first_key]['strokes_count'] == 2
        assert stats['input_providers'][first_key]['percent'] == 100.0
        assert len(stats['properties']) == 2
        assert len(stats['sensor_channels']) == 4
        assert len(stats['document_bounds']) == 6
        assert stats['knowledge_graph']['statements_count'] == 6
        assert len(stats['views']) == 1
        assert stats['uim_version'] == '3.1.0'

    except Exception as e:
        pytest.fail(f"StatisticsAnalyzer should not raise an exception: {e}")
    # The model should have a knowledge graph and a tree
    assert artificial_model.has_knowledge_graph()
    # The model should have a tree
    assert artificial_model.has_tree(CommonViews.HWR_VIEW.value)
    words, entities, text = uim_extract_text_and_semantics_from(artificial_model, CommonViews.HWR_VIEW.value)
    assert len(words) > 0
    assert len(entities) == 0
    assert len(text) > 0
    assert text == '--'
    for word in words:
        assert word['text'] == '-'


@pytest.mark.parametrize('path', uim_310_files() + uim_300_files())
def test_uim_analyzer(path: Path):
    """
    Test the statistics analyzer.
    """
    parser: UIMParser = UIMParser()
    ink_model: InkModel = parser.parse(path)
    try:
        model_analyser: StatisticsAnalyzer = StatisticsAnalyzer()
        stats: Dict[str, Any] = model_analyser.analyze(ink_model)
        assert len(stats) > 0
    except StatisticsError as e:
        if len(ink_model.strokes) > 1:
            pytest.fail(f"StatisticsAnalyzer should not raise a StatisticsError: {e}")
        # Single stroke models are expected to raise a StatisticsError


def test_analyzer_to_few_strokes():
    """
    Test the statistics analyzer.
    """
    artificial_model: InkModel = created_model(num_point=10, number_strokes=1, with_timestamps=False,
                                               with_pressure=False)
    with pytest.raises(StatisticsError):
        model_analyser: StatisticsAnalyzer = StatisticsAnalyzer()
        model_analyser.analyze(artificial_model)


def test_analyzer_to_no_pressure():
    """
    Test the statistics analyzer.
    """
    artificial_model: InkModel = created_model(num_point=10, number_strokes=2, with_pressure=False)
    model_analyser: StatisticsAnalyzer = StatisticsAnalyzer()
    stats = model_analyser.analyze(artificial_model)


def test_analyzer_as_strided_array():
    """
    Test the statistics analyzer.
    """
    num_points: int = 100
    artificial_model: InkModel = created_model(num_point=num_points, number_strokes=2, with_pressure=False)
    pressure_nans = lambda lst: all(math.isnan(lst[i]) for i in range(3, len(lst), 4))
    pressure_zero = lambda lst: all(lst[i] == 0. for i in range(3, len(lst), 4))
    timestamps_nans = lambda lst: all(math.isnan(lst[i]) for i in range(2, len(lst), 4))
    timestamps_zero = lambda lst: all(lst[i] == 0. for i in range(2, len(lst), 4))
    for stroke in artificial_model.strokes:
        ref = as_strided_array(artificial_model, stroke=stroke,
                               handle_missing_data=HandleMissingDataPolicy.SKIP_STROKE)
        assert ref is None
        with pytest.raises(ValueError):
            as_strided_array(artificial_model, stroke=stroke,
                             handle_missing_data=HandleMissingDataPolicy.THROW_EXCEPTION)
        ref = as_strided_array(artificial_model, stroke=stroke,
                               handle_missing_data=HandleMissingDataPolicy.FILL_WITH_NAN)
        assert ref is not None
        assert len(ref) == num_points * 4  # X, Y, Size, Pressure
        assert pressure_nans(ref)
        ref = as_strided_array(artificial_model, stroke=stroke,
                               handle_missing_data=HandleMissingDataPolicy.FILL_WITH_ZEROS)
        assert ref is not None
        assert len(ref) == num_points * 4
        assert pressure_zero(ref)
    artificial_model: InkModel = created_model(num_point=num_points, number_strokes=2, with_timestamps=False)
    for stroke in artificial_model.strokes:
        ref = as_strided_array(artificial_model, stroke=stroke,
                               handle_missing_data=HandleMissingDataPolicy.SKIP_STROKE)
        assert ref is None
        with pytest.raises(ValueError):
            as_strided_array(artificial_model, stroke=stroke,
                             handle_missing_data=HandleMissingDataPolicy.THROW_EXCEPTION)
        ref = as_strided_array(artificial_model, stroke=stroke,
                               handle_missing_data=HandleMissingDataPolicy.FILL_WITH_NAN)
        assert ref is not None
        assert len(ref) == num_points * 4  # X, Y, Size, Pressure
        assert timestamps_nans(ref)
        ref = as_strided_array(artificial_model, stroke=stroke,
                               handle_missing_data=HandleMissingDataPolicy.FILL_WITH_ZEROS)
        assert ref is not None
        assert len(ref) == num_points * 4
        assert timestamps_zero(ref)
    artificial_model: InkModel = created_model(num_point=num_points, number_strokes=2, with_pressure=False,
                                               with_timestamps=False, add_sensor_data=False)
    for stroke in artificial_model.strokes:
        ref = as_strided_array(artificial_model, stroke=stroke,
                               handle_missing_data=HandleMissingDataPolicy.FILL_WITH_ZEROS)
        assert timestamps_zero(ref)
        assert pressure_zero(ref)





# -*- coding: utf-8 -*-
# Copyright © 2021-present Wacom Authors. All Rights Reserved.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
from io import BytesIO
from pathlib import Path

import bitstring
import pytest

from uim.codec.base import ContentType, CompressionType
from uim.codec.context.decoder import DecoderContext
from uim.codec.context.encoder import EncoderContext
from uim.codec.parser.base import SupportedFormats, FormatException
from uim.codec.parser.uim import UIMParser
from uim.codec.writer.encoder.encoder_3_1_0 import UIMEncoder310
from uim.model import UUIDIdentifier
from uim.model.ink import InkModel
from uim.model.inkinput.sensordata import ChannelData
from uim.model.semantics.schema import CommonViews

# Test data directory
test_data_dir: Path = Path(__file__).parent / '../ink/uim_3.1.0/'
raster_data_dir: Path = Path(__file__).parent / '../ink/special'


def uim_files() -> list:
    return [f for f in test_data_dir.iterdir() if f.is_file() and f.name.endswith('.uim')]


@pytest.mark.parametrize('path', uim_files())
def test_uim_3_1_0_bytes_io(path: Path):
    with path.open('rb') as fp:
        uim_bytes: bytes = fp.read()
        parser: UIMParser = UIMParser()
        ink_model: InkModel = parser.parse(BytesIO(uim_bytes))
        assert ink_model.has_brushes()
        assert ink_model.has_ink_data()
        assert ink_model.has_ink_structure()
        assert len(ink_model.strokes) > 0


@pytest.mark.parametrize('path', uim_files())
def test_uim_3_1_0(path: Path):
    parser: UIMParser = UIMParser()
    ink_model: InkModel = parser.parse(path)
    assert ink_model.has_brushes()
    assert ink_model.has_ink_data()
    assert ink_model.has_ink_structure()
    assert len(ink_model.strokes) > 0
    # Validate encoding
    uim_bytes: bytes = UIMEncoder310().encode(ink_model)
    assert len(uim_bytes) > 0
    # Validate decoding
    ink_model_decoded: InkModel = parser.parse(uim_bytes)
    assert ink_model_decoded.has_brushes()
    assert ink_model_decoded.has_ink_data()
    assert ink_model_decoded.has_ink_structure()
    assert len(ink_model_decoded.strokes) > 0
    # Validate sensor data
    for sd_1, sd_2 in zip(ink_model.sensor_data.sensor_data, ink_model_decoded.sensor_data.sensor_data):
        for s_idx in range(len(sd_1.data_channels)):
            s1_data: ChannelData = sd_1.data_channels[s_idx]
            s2_data: ChannelData = sd_2.data_channels[s_idx]
            # Ids must be the same
            assert s1_data.id == s2_data.id
            for v_idx, val_1 in enumerate(s1_data.values):
                val_2 = sd_2.data_channels[s_idx].values[v_idx]
                assert val_1 == pytest.approx(val_2)
    for str_1, str_2 in zip(ink_model.strokes, ink_model_decoded.strokes):
        # Ids must be the same
        assert str_1.id == str_2.id
    assert str(ink_model) == str(ink_model_decoded)


def test_load_raster():
    """
    Test loading raster data.
    """
    raster_file: Path = raster_data_dir / 'raster.uim'
    parser: UIMParser = UIMParser()
    ink_model: InkModel = parser.parse(raster_file)
    assert ink_model.has_brushes()
    assert len(ink_model.brushes.raster_brushes) == 4
    for raster in ink_model.brushes.raster_brushes:
        assert str(raster)


def test_not_accept():
    """
    Testing type not accepted.
    """
    parser: UIMParser = UIMParser()
    with pytest.raises(TypeError):
        parser.parse(45)


def test_wrong_bytes():
    """
    Test wrong bytes.
    """
    parser: UIMParser = UIMParser()
    with pytest.raises(FormatException):
        parser.parse(BytesIO(b'Hello Ink'))


def test_wrong_version():
    """
    Test wrong version.
    """
    reserved: bytes = b'\x00'
    buffer: bitstring.BitStream = bitstring.BitStream()
    buffer.append(b'RIFF')
    # Size (uint32): size of all chunks (4 Bytes)
    buffer.append(int(0).to_bytes(4, byteorder="little"))
    buffer.append(b'UINK')
    buffer.append(b'HEAD')
    buffer.append(UIMEncoder310.VERSION_MAJOR)
    buffer.append(UIMEncoder310.VERSION_MINOR)
    buffer.append(UIMEncoder310.VERSION_MINOR)
    buffer.append(ContentType.PROTOBUF.value)
    buffer.append(CompressionType.NONE.value)
    buffer.append(reserved)
    buffer.append(reserved)
    buffer.append(reserved)
    parser: UIMParser = UIMParser()
    with pytest.raises(FormatException):
        parser.parse(BytesIO(buffer.tobytes()))


def test_missing_header():
    """
    Test missing header.
    """
    reserved: bytes = b'\x00'
    buffer: bitstring.BitStream = bitstring.BitStream()
    buffer.append(b'RIFF')
    # Size (uint32): size of all chunks (4 Bytes)
    buffer.append(int(0).to_bytes(4, byteorder="little"))
    buffer.append(b'UINK')
    buffer.append(UIMEncoder310.VERSION_MAJOR)
    buffer.append(UIMEncoder310.VERSION_MINOR)
    buffer.append(UIMEncoder310.VERSION_MINOR)
    buffer.append(ContentType.PROTOBUF.value)
    buffer.append(CompressionType.NONE.value)
    buffer.append(reserved)
    buffer.append(reserved)
    buffer.append(reserved)
    parser: UIMParser = UIMParser()
    with pytest.raises(FormatException):
        parser.parse(BytesIO(buffer.tobytes()))


def test_encoder_context():
    """
    Test encoder context.
    """
    assert (EncoderContext.view_name(CommonViews.HWR_VIEW.value, SupportedFormats.UIM_VERSION_3_1_0)
            == CommonViews.HWR_VIEW.value)
    assert (EncoderContext.view_name(CommonViews.LEGACY_HWR_VIEW.value, SupportedFormats.UIM_VERSION_3_1_0)
            == CommonViews.HWR_VIEW.value)
    assert (EncoderContext.view_name(CommonViews.NER_VIEW.value, SupportedFormats.UIM_VERSION_3_1_0)
            == CommonViews.NER_VIEW.value)
    assert (EncoderContext.view_name(CommonViews.LEGACY_NER_VIEW.value, SupportedFormats.UIM_VERSION_3_1_0)
            == CommonViews.NER_VIEW.value)
    assert (EncoderContext.view_name("Undefined", SupportedFormats.UIM_VERSION_3_1_0) == "Undefined")
    with pytest.raises(FormatException):
        # Only UIM 3.1.0 is supported
        EncoderContext.view_name(CommonViews.HWR_VIEW.value, SupportedFormats.UIM_VERSION_3_0_0)

    ink_model: InkModel = InkModel()
    context = EncoderContext(version=SupportedFormats.UIM_VERSION_3_1_0.value, ink_model=ink_model)
    assert context.format_version == SupportedFormats.UIM_VERSION_3_1_0.value


def test_decoder_context():
    """
    Test decoder context.
    """
    ink_model: InkModel = InkModel()
    context = DecoderContext(version=SupportedFormats.UIM_VERSION_3_1_0.value, ink_model=ink_model)
    assert context.format_version == SupportedFormats.UIM_VERSION_3_1_0.value
    with pytest.raises(ValueError):
        context.stroke_by_identifier(UUIDIdentifier.id_generator().hex)



# -*- coding: utf-8 -*-
# Copyright © 2024-present Wacom Authors. All Rights Reserved.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
from pathlib import Path
from typing import List, Dict, Any, Tuple

from uim.codec.parser.uim import UIMParser
from uim.model.helpers.schema_content_extractor import uim_schema_semantics_from
from uim.model.ink import InkModel
from uim.model.semantics.schema import CommonViews


def build_tree(node_list: List[Dict[str, Any]]):
    """
    Build a tree structure from the node list.
    Parameters
    ----------
    node_list: `List[Dict[str, Any]]`
        List of nodes
    """
    # Step 1: Create dictionaries for nodes and parent-child relationships
    children_dict: Dict[str, Any] = {}

    for node in node_list:
        parent_uri: str = node['parent_uri']
        if parent_uri is not None:
            if parent_uri not in children_dict:
                children_dict[parent_uri] = []
            children_dict[parent_uri].append(node)

    # Step 2: Define a recursive function to print the tree
    def print_tree(node_display: Dict[str, Any], indent: int = 0):
        info: str = ""
        attributes: List[Tuple[str, Any]] = node_display.get('attributes', [])
        if "path_id" in node_display:
            info = f"(#strokes:={len(node_display['path_id'])})"
        elif "bounding_box" in node_display:
            info = (f"(x:={node_display['bounding_box']['x']}, y:={node_display['bounding_box']['y']}, "
                    f"width:={node_display['bounding_box']['width']}, "
                    f"height:={node_display['bounding_box']['height']})")
        print('|' + '-' * indent + f" [type:={node_display['type']}] - {info}")
        if len(attributes) > 0:
            print('|' + ' ' * (indent + 4) + "| -[Attributes:]")
            for key, value in attributes:
                print('|' + ' ' * (indent + 8) + f"\t|-- {key}:={value}")
        if node_display['node_uri'] in children_dict:
            for child in children_dict[node_display['node_uri']]:
                print_tree(child, indent + 4)

    # Step 3: Find the root node (where parent_uri is None) and start printing the tree
    for node in node_list:
        if node['parent_uri'] is None:
            print_tree(node)


if __name__ == '__main__':
    parser: UIMParser = UIMParser()
    # Parse UIM v3.0.0
    ink_model: InkModel = parser.parse(Path(__file__).parent / '..' / 'ink' / 'schemas' / 'math-structures.uim')
    math_structures: List[Dict[str, Any]] = uim_schema_semantics_from(ink_model,
                                                                      semantic_view=CommonViews.HWR_VIEW.value)
    # Print the tree structure
    build_tree(math_structures)



# -*- coding: utf-8 -*-
# Copyright © 2021 Wacom Authors. All Rights Reserved.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
import csv
import uuid
from collections import defaultdict
from pathlib import Path
from typing import List, Dict

from uim.codec.parser.base import SupportedFormats
from uim.codec.writer.encoder.encoder_3_1_0 import UIMEncoder310
from uim.model.base import UUIDIdentifier
from uim.model.helpers.serialize import json_encode
from uim.model.ink import InkModel, InkTree, ViewTree
from uim.model.inkdata.brush import VectorBrush, BrushPolygon, BrushPolygonUri
from uim.model.inkdata.strokes import Spline, Style, Stroke, LayoutMask
from uim.model.inkinput.inputdata import Environment, InkInputProvider, InkInputType, InputDevice, SensorChannel, \
    InkSensorType, InkSensorMetricType, SensorChannelsContext, SensorContext, InputContext, unit2unit, Unit
from uim.model.inkinput.sensordata import SensorData, InkState
from uim.model.semantics import schema
from uim.model.semantics.node import StrokeGroupNode, StrokeNode, URIBuilder
from uim.utils.matrix import Matrix4x4


def create_sensor_data(data_collection: Dict[str, List[float]],
                       input_context_id: uuid.UUID, channels: List[SensorChannel]) -> SensorData:
    """
    Create sensor data from a data collection.
    Parameters
    ----------
    data_collection: Dict[str, List[float]]

    input_context_id
    channels

    Returns
    -------
    SensorData
        Instance of SensorData
    """
    sd: SensorData = SensorData(UUIDIdentifier.id_generator(), input_context_id=input_context_id, state=InkState.PLANE)
    sd.add_data(channels[0], [unit2unit(Unit.DIP, Unit.M, v) for v in data_collection['SPLINE_X']])
    sd.add_data(channels[1], [unit2unit(Unit.DIP, Unit.M, v) for v in data_collection['SPLINE_Y']])
    sd.add_timestamp_data(channels[2], data_collection['SENSOR_TIMESTAMP'])
    sd.add_data(channels[3], data_collection['SENSOR_PRESSURE'])
    sd.add_data(channels[4], data_collection['SENSOR_AZIMUTH'])
    sd.add_data(channels[5], data_collection['SENSOR_ALTITUDE'])
    return sd


def load_sensor_data(csv_path: Path, input_context_id: uuid.UUID, channels: List[SensorChannel]) -> List[SensorData]:
    """
    Load sensor data from a CSV file.

    Parameters
    ----------
    csv_path: Path
        Path to the CSV file
    input_context_id: uuid.UUID
        Input context ID
    channels: List[SensorChannel]
        List of sensor channels

    Returns
    -------
    List[SensorData]
        List of sensor data
    """
    sensor_data_values: List[SensorData] = []
    data_collection: Dict[str, List[float]] = defaultdict(list)

    with csv_path.open('r') as f:
        reader = csv.reader(f)
        header: List[str] = next(reader)
        if header != ['idx', 'SPLINE_X', 'SPLINE_Y',
                      'SENSOR_TIMESTAMP', 'SENSOR_PRESSURE', 'SENSOR_ALTITUDE', 'SENSOR_AZIMUTH']:
            raise ValueError("Invalid CSV file format")
        last_idx: int = 0
        for row in reader:
            row_idx: int = int(row[0])
            if row_idx != last_idx:
                sensor_data_values.append(create_sensor_data(data_collection, input_context_id, channels))
                data_collection.clear()
            for idx, value in enumerate(row[1:], start=1):
                data_collection[header[idx]].append(float(value))
            last_idx = row_idx
        if len(data_collection) > 0:
            sensor_data_values.append(create_sensor_data(data_collection, input_context_id, channels))
    return sensor_data_values


def create_strokes(sensor_data_items: List[SensorData], style_stroke: Style, x_id: uuid.UUID, y_id: uuid.UUID) \
        -> List[Stroke]:
    """
    Create strokes from sensor data.

    Parameters
    ----------
    sensor_data_items: List[SensorData]
        List of sensor data
    style_stroke: Style
        Style of the stroke
    x_id: uuid.UUID
        Reference id of x sensor channel
    y_id: uuid.UUID
        Reference id of y sensor channel

    Returns
    -------
    List[Stroke]
        List of strokes
    """
    stroke_items: List[Stroke] = []
    for sensor_data_i in sensor_data_items:
        path: List[float] = []
        # The spline path contains x, y values
        mask: int = LayoutMask.X.value | LayoutMask.Y.value
        for x, y in zip(sensor_data_i.get_data_by_id(x_id).values, sensor_data_i.get_data_by_id(y_id).values):
            path.append(unit2unit(Unit.M, Unit.DIP, x))
            path.append(unit2unit(Unit.M, Unit.DIP, y))

        spline: Spline = Spline(layout_mask=mask, data=path)
        # Create a stroke from spline
        s_i: Stroke = Stroke(sid=UUIDIdentifier.id_generator(), spline=spline, style=style_stroke)
        stroke_items.append(s_i)
    return stroke_items


if __name__ == '__main__':
    # Creates an ink model from the scratch.
    ink_model: InkModel = InkModel(version=SupportedFormats.UIM_VERSION_3_1_0.value)
    # Setting a unit scale factor
    ink_model.unit_scale_factor = 1.5
    # Using a 4x4 matrix for scaling
    ink_model.transform = Matrix4x4.create_scale(1.5)

    # Properties are added as key-value pairs
    ink_model.properties.append(("Author", "Markus"))
    ink_model.properties.append(("Locale", "en_US"))

    # Create an environment
    env: Environment = Environment()
    # This should describe the environment in which the ink was captured
    env.properties.append(("wacom.ink.sdk.lang", "js"))
    env.properties.append(("wacom.ink.sdk.version", "2.0.0"))
    env.properties.append(("runtime.type", "WEB"))
    env.properties.append(("user.agent.brands", "Chromium 126, Google Chrome 126"))
    env.properties.append(("user.agent.platform", "macOS"))
    env.properties.append(("user.agent.mobile", "false"))
    env.properties.append(("app.id", "sample_create_model_vector"))
    env.properties.append(("app.version", "1.0.0"))
    ink_model.input_configuration.environments.append(env)

    # Ink input provider can be pen, mouse or touch.
    provider: InkInputProvider = InkInputProvider(input_type=InkInputType.PEN)
    ink_model.input_configuration.ink_input_providers.append(provider)

    # Input device is the sensor (pen tablet, screen, etc.)
    input_device: InputDevice = InputDevice()
    input_device.properties.append(("dev.manufacturer", "Wacom"))
    input_device.properties.append(("dev.model", "Wacom One"))
    input_device.properties.append(("dev.product.code", "DTC-133"))
    input_device.properties.append(("dev.graphics.resolution", "1920x1080"))
    ink_model.input_configuration.devices.append(input_device)
    # Create a group of sensor channels
    sensor_channels: list = [
        SensorChannel(channel_type=InkSensorType.X, metric=InkSensorMetricType.LENGTH, resolution=1.0,
                      ink_input_provider_id=provider.id, input_device_id=input_device.id),
        SensorChannel(channel_type=InkSensorType.Y, metric=InkSensorMetricType.LENGTH, resolution=1.0,
                      ink_input_provider_id=provider.id,  input_device_id=input_device.id),
        SensorChannel(channel_type=InkSensorType.TIMESTAMP, metric=InkSensorMetricType.TIME, resolution=1000.0,
                      precision=0,
                      ink_input_provider_id=provider.id, input_device_id=input_device.id),
        SensorChannel(channel_type=InkSensorType.PRESSURE, metric=InkSensorMetricType.NORMALIZED, resolution=1.0,
                      channel_min=0., channel_max=1.0,
                      ink_input_provider_id=provider.id, input_device_id=input_device.id),
        SensorChannel(channel_type=InkSensorType.ALTITUDE, metric=InkSensorMetricType.ANGLE, resolution=1.0,
                      channel_min=0., channel_max=1.5707963705062866,
                      ink_input_provider_id=provider.id, input_device_id=input_device.id),
        SensorChannel(channel_type=InkSensorType.AZIMUTH, metric=InkSensorMetricType.ANGLE, resolution=1.0,
                      channel_min=-3.1415927410125732, channel_max=3.1415927410125732,
                      ink_input_provider_id=provider.id, input_device_id=input_device.id)
    ]
    # Create a sensor channels context
    scc_wacom_one: SensorChannelsContext = SensorChannelsContext(channels=sensor_channels,
                                                                 ink_input_provider_id=provider.id,
                                                                 input_device_id=input_device.id,
                                                                 latency=0,
                                                                 sampling_rate_hint=240)

    # Add sensor channel contexts
    sensor_context: SensorContext = SensorContext()
    sensor_context.add_sensor_channels_context(scc_wacom_one)
    ink_model.input_configuration.sensor_contexts.append(sensor_context)

    # Create the input context using the Environment and the Sensor Context
    input_context: InputContext = InputContext(environment_id=env.id, sensor_context_id=sensor_context.id)
    ink_model.input_configuration.input_contexts.append(input_context)

    # Create sensor data
    # The CSV file contains sensor data for strokes
    # idx,SPLINE_X,SPLINE_Y,SENSOR_TIMESTAMP,SENSOR_PRESSURE,SENSOR_ALTITUDE,SENSOR_AZIMUTH
    sensor_data = load_sensor_data(Path(__file__).parent / '..' / 'ink' / 'sensor_data' / 'ink.csv', input_context.id,
                                   sensor_channels)
    # Add sensor data to the model
    for sensor_data_i in sensor_data:
        ink_model.sensor_data.add(sensor_data_i)

    # We need to define a brush polygon
    points: list = [(10, 10), (0, 10), (0, 0), (10, 0)]
    brush_polygons: list = [BrushPolygon(min_scale=0., points=points)]

    # Create the brush object using polygons
    vector_brush_0: VectorBrush = VectorBrush(
        "app://qa-test-app/vector-brush/MyTriangleBrush",
        brush_polygons)

    # Add it to the model
    ink_model.brushes.add_vector_brush(vector_brush_0)

    # Add a brush specified with shape Uris
    poly_uris: list = [
        BrushPolygonUri("will://brush/3.0/shape/Circle?precision=20&radius=1", 0.),
        BrushPolygonUri("will://brush/3.0/shape/Ellipse?precision=20&radiusX=1&radiusY=0.5", 4.0)
    ]
    # Define a second brush
    vector_brush_1: VectorBrush = VectorBrush(
        "app://qa-test-app/vector-brush/MyEllipticBrush",
        poly_uris)
    # Add it to the model
    ink_model.brushes.add_vector_brush(vector_brush_1)

    # Specify the layout of the stroke data, in this case the stroke will have variable X, Y and Size properties.
    layout_mask: int = LayoutMask.X.value | LayoutMask.Y.value | LayoutMask.SIZE.value

    # Create some style
    style: Style = Style(brush_uri=vector_brush_1.name)
    # Set the color of the strokes
    style.path_point_properties.red = 0.1
    style.path_point_properties.green = 0.2
    style.path_point_properties.blue = 0.4
    style.path_point_properties.alpha = 1.0

    # Create the strokes
    strokes = create_strokes(sensor_data, style, sensor_channels[0].id, sensor_channels[1].id)
    # First you need a root group to contain the strokes
    root: StrokeGroupNode = StrokeGroupNode(UUIDIdentifier.id_generator())

    # Assign the group as the root of the main ink tree
    ink_model.ink_tree = InkTree()
    ink_model.ink_tree.root = root

    # Adding the strokes to the root group
    for stroke in strokes:
        root.add(StrokeNode(stroke))

    # Adding view for handwriting recognition results
    hwr_tree: ViewTree = ViewTree(schema.CommonViews.HWR_VIEW.value)
    # Add view right after creation, to avoid warnings that tree is not yet attached
    ink_model.add_view(hwr_tree)
    # Create a root node for the HWR view
    hwr_root: StrokeGroupNode = StrokeGroupNode(UUIDIdentifier.id_generator())
    hwr_tree.root = hwr_root
    ink_model.knowledge_graph.append(schema.SemanticTriple(hwr_root.uri, schema.IS, schema.SegmentationSchema.ROOT))
    ink_model.knowledge_graph.append(schema.SemanticTriple(hwr_root.uri, schema.SegmentationSchema.HAS_REPRESENTS_VIEW,
                                                           schema.CommonViews.HWR_VIEW.value))

    # Here you can add the same strokes as in the main tree, but you can organize them in a different way
    # (put them in different groups)
    # You are not supposed to add strokes that are not already in the main tree.
    text_region: StrokeGroupNode = StrokeGroupNode(UUIDIdentifier.id_generator())
    hwr_root.add(text_region)
    ink_model.knowledge_graph.append(schema.SemanticTriple(text_region.uri, schema.IS,
                                                           schema.SegmentationSchema.TEXT_REGION))

    # The text_line root denotes the text line
    text_line: StrokeGroupNode = StrokeGroupNode(UUIDIdentifier.id_generator())
    text_region.add(text_line)
    ink_model.knowledge_graph.append(schema.SemanticTriple(text_line.uri, schema.IS,
                                                           schema.SegmentationSchema.TEXT_LINE))

    # The word node denotes a word
    word: StrokeGroupNode = StrokeGroupNode(UUIDIdentifier.id_generator())
    text_line.add(word)
    ink_model.knowledge_graph.append(schema.SemanticTriple(word.uri, schema.IS, schema.SegmentationSchema.WORD))
    ink_model.knowledge_graph.append(schema.SemanticTriple(word.uri, schema.SegmentationSchema.HAS_CONTENT, "ink"))
    ink_model.knowledge_graph.append(schema.SemanticTriple(word.uri, schema.SegmentationSchema.HAS_LANGUAGE, "en_US"))

    # Add the strokes to the word
    for stroke_i in strokes:
        word.add(StrokeNode(stroke_i))

    # We need a URI builder
    uri_builder: URIBuilder = URIBuilder()

    # Create a named entity
    named_entity_uri: str = uri_builder.build_named_entity_uri(UUIDIdentifier.id_generator())
    ink_model.knowledge_graph.append(schema.SemanticTriple(word.uri,
                                                           schema.NamedEntityRecognitionSchema.HAS_PART,
                                                           named_entity_uri))

    # Add knowledge for the named entity
    ink_model.knowledge_graph.append(schema.SemanticTriple(named_entity_uri, "hasPart-0", word.uri))
    ink_model.knowledge_graph.append(schema.SemanticTriple(named_entity_uri,
                                                           schema.NamedEntityRecognitionSchema.HAS_LABEL, "Ink"))
    ink_model.knowledge_graph.append(schema.SemanticTriple(named_entity_uri,
                                                           schema.NamedEntityRecognitionSchema.HAS_LANGUAGE, "en_US"))
    ink_model.knowledge_graph.append(schema.SemanticTriple(named_entity_uri,
                                                           schema.NamedEntityRecognitionSchema.HAS_CONFIDENCE, "0.95"))
    ink_model.knowledge_graph.append(schema.SemanticTriple(named_entity_uri,
                                                           schema.NamedEntityRecognitionSchema.HAS_ARTICLE_URL,
                                                       'https://en.wikipedia.org/wiki/Ink'))
    ink_model.knowledge_graph.append(schema.SemanticTriple(named_entity_uri,
                                                           schema.NamedEntityRecognitionSchema.HAS_UNIQUE_ID,
                                                           'Q127418'))
    # Save the model, this will overwrite an existing file
    with open('3_1_0_vector.uim', 'wb') as uim:
        # unicode(data) auto-decodes data to unicode if str
        uim.write(UIMEncoder310().encode(ink_model))
    # Convert the model to JSON
    with open('ink.json', 'w') as f:
        # json_encode is a helper function to convert the model to JSON
        f.write(json_encode(ink_model))



# -*- coding: utf-8 -*-
# Copyright © 2021-present Wacom Authors. All Rights Reserved.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
import uuid
from typing import List, Tuple

import pytest

from uim.codec.parser.base import SupportedFormats
from uim.codec.parser.uim import UIMParser
from uim.codec.writer.encoder.encoder_3_1_0 import UIMEncoder310
from uim.model import UUIDIdentifier
from uim.model.base import InkModelException, Identifier, IdentifiableMethod
from uim.model.helpers.treeiterator import PreOrderEnumerator
from uim.model.ink import InkModel, ViewTree, InkTree
from uim.model.inkdata.brush import VectorBrush, BrushPolygon, BrushPolygonUri, RasterBrush, RotationMode, BlendMode, \
    BlendModeURIs
from uim.model.inkdata.strokes import Spline, Style, Stroke, LayoutMask, PathPointProperties
from uim.model.inkinput.inputdata import Environment, InkInputProvider, InkInputType, InputDevice, SensorChannel, \
    InkSensorType, InkSensorMetricType, SensorChannelsContext, SensorContext, InputContext
from uim.model.inkinput.sensordata import SensorData, InkState
from uim.model.semantics import schema
from uim.model.semantics.node import StrokeGroupNode, StrokeNode, StrokeFragment, URIBuilder, InkNode
from uim.model.semantics.schema import CommonViews, SemanticTriple, DOCUMENT_AUTHOR_OBJECT, DOCUMENT_LOCALE_OBJECT
from uim.model.semantics.structures import BoundingBox
from uim.utils.matrix import Matrix4x4


def create_ink_model(with_sensor_data: bool = True) -> InkModel:
    # Create the model
    ink_model: InkModel = InkModel(version=SupportedFormats.UIM_VERSION_3_1_0.value)
    # Setting a unit scale factor
    ink_model.unit_scale_factor = 1.5
    # Using a 4x4 matrix for scaling
    ink_model.transform = Matrix4x4.create_scale(1.5)
    # Properties are added as key-value pairs
    ink_model.properties.append((DOCUMENT_AUTHOR_OBJECT, "John"))
    ink_model.properties.append((DOCUMENT_LOCALE_OBJECT, "en_US"))
    ink_model.properties.append(("OtherStuff", "Yes"))
    # Create an environment
    env: Environment = Environment()
    env.properties.append(("env.name", "My Environment"))
    env.properties.append(("os.id", "98765"))
    env.properties.append(("os.name", "Windows"))
    env.properties.append(("os.version", "10.0.18362.239"))
    env.properties.append(("os.build", "239"))
    env.properties.append(("os.platform", "whatever"))
    env.add_environment_property("app.id", "123456")
    assert str(env)
    assert env != ink_model
    ink_model.input_configuration.environments.append(env)
    # Ink input provider can be pen, mouse or touch.
    provider: InkInputProvider = InkInputProvider(input_type=InkInputType.MOUSE)
    provider.properties.append(("dev.id", "1234567"))
    assert provider != env
    ink_model.input_configuration.ink_input_providers.append(provider)
    # Input device is the sensor (pen tablet, screen, etc.)
    input_device: InputDevice = InputDevice()
    input_device.properties.append(("dev.id", "123454321"))
    input_device.properties.append(("dev.manufacturer", "Wacom"))
    ink_model.input_configuration.devices.append(input_device)
    # Create a group of sensor channels
    sensor_channels_tablet: list = [
        SensorChannel(channel_type=InkSensorType.X, metric=InkSensorMetricType.LENGTH, resolution=1.0),
        SensorChannel(channel_type=InkSensorType.Y, metric=InkSensorMetricType.LENGTH, resolution=1.0),
        SensorChannel(channel_type=InkSensorType.TIMESTAMP, metric=InkSensorMetricType.TIME, resolution=1.0),
        SensorChannel(channel_type=InkSensorType.PRESSURE, metric=InkSensorMetricType.FORCE, resolution=1.0),
        SensorChannel(channel_type=InkSensorType.AZIMUTH, metric=InkSensorMetricType.ANGLE, resolution=1.0),
        SensorChannel(channel_type=InkSensorType.ALTITUDE, metric=InkSensorMetricType.ANGLE, resolution=1.0)
    ]
    assert sensor_channels_tablet[0] != sensor_channels_tablet[1]
    assert sensor_channels_tablet[0] != ink_model
    # Create a sensor channel context
    scc_tablet: SensorChannelsContext = SensorChannelsContext(channels=sensor_channels_tablet,
                                                              ink_input_provider_id=provider.id,
                                                              input_device_id=input_device.id)
    # We can create an additional input device, for example one providing pressure via Bluetooth
    scc_bluetooth, _ = pen_sensor_context(ink_model, provider)
    # Add all sensor channel contexts
    sensor_context: SensorContext = SensorContext()
    sensor_context.add_sensor_channels_context(scc_tablet)
    sensor_context.add_sensor_channels_context(scc_bluetooth)
    ink_model.input_configuration.sensor_contexts.append(sensor_context)
    # Create the input context using the Environment and the Sensor Context
    input_context: InputContext = InputContext(environment_id=env.id, sensor_context_id=sensor_context.id)
    ink_model.input_configuration.input_contexts.append(input_context)
    sensor_data_id: uuid.UUID = UUIDIdentifier.id_generator()
    if with_sensor_data:
        # Create sensor data
        sensor_data_0: SensorData = SensorData(sensor_data_id, input_context_id=input_context.id,
                                               state=InkState.PLANE)
        sensor_data_0.add_timestamp_data(sensor_channels_tablet[2], [0., 0.1, 0.2, 0.3, 0.4])
        sensor_data_0.add_data(sensor_channels_tablet[0], [100.4, 103.7, 110.1])
        sensor_data_0.add_data(sensor_channels_tablet[1], [200.1, 202.0, 207.0])
        sensor_data_0.add_data(sensor_channels_tablet[3], [0.1, 0.4, 0.2])
        sensor_data_0.add_data(sensor_channels_tablet[4], [50., 51.0, 52.0])
        sensor_data_0.add_data(sensor_channels_tablet[5], [50., 51.0, 52.0])

        # Add sensor data to the model
        ink_model.sensor_data.add(sensor_data_0)
    # We need to define a brush polygon
    points: list = [(10, 10), (0, 10), (0, 0)]
    brush_polygons: list = [BrushPolygon(min_scale=0., points=points)]
    # PathPointProperties
    path_point_properties = PathPointProperties(
        size=1.0,
        red=0.0,
        green=1.0,
        blue=0.0,
        alpha=1.0,
        rotation=0.0,
        scale_x=1.0,
        scale_y=1.0,
        scale_z=1.0,
        offset_x=0.0,
        offset_y=0.0,
        offset_z=0.0
    )
    # Create the brush object using polygons
    vector_brush_0: VectorBrush = VectorBrush("app://qa-test-app/vector-brush/MyTriangleBrush", brush_polygons)
    # Add it to the model
    ink_model.brushes.add_vector_brush(vector_brush_0)
    # Add a brush specified with shape Uris
    poly_uris: list = [
        BrushPolygonUri("will://brush/3.0/shape/Circle?precision=20&radius=1", 0.),
        BrushPolygonUri("will://brush/3.0/shape/Ellipse?precision=20&radiusX=1&radiusY=0.5", 4.0)
    ]
    vector_brush_1: VectorBrush = VectorBrush(
        "app://qa-test-app/vector-brush/MyEllipticBrush",
        poly_uris)
    raster_brush_0: RasterBrush = RasterBrush(
        name="app://qa-test-app/raster-brush/MyRasterBrush",
        spacing=10., scattering=5., rotation=RotationMode.TRAJECTORY, shape_textures=[bytes([10, 20]),
                                                                                      bytes([30, 20])],
        fill_width=2.0, fill_height=0.3,
        fill_texture=bytes([10, 10, 20, 15, 17, 20, 25, 16, 34, 255, 23, 0, 34, 255, 23, 255]),
        randomize_fill=False, blend_mode=BlendMode.SOURCE_OVER)
    assert input_context != raster_brush_0
    # Add it to the model
    ink_model.brushes.add_raster_brush(raster_brush_0)
    raster_brush_1: RasterBrush = RasterBrush(
        name="app://qa-test-app/raster-brush/MyRasterBrush1",
        spacing=10.0, scattering=5., rotation=RotationMode.TRAJECTORY, fill_width=2., fill_height=0.3,
        fill_texture_uri="app://qa-test-app/raster-brush-fill/mixedShapesGL",
        shape_texture_uris=[
            "app://qa-test-app/raster-brush-shape/mixedShapesGL_128x128",
            "app://qa-test-app/raster-brush-shape/mixedShapesGL_64x64",
            "app://qa-test-app/raster-brush-shape/mixedShapesGL_32x32",
            "app://qa-test-app/raster-brush-shape/mixedShapesGL_16x16"
        ], randomize_fill=False, blend_mode=BlendMode.SOURCE_OVER)
    ink_model.brushes.add_raster_brush(raster_brush_1)
    ink_model.brushes.add_vector_brush(vector_brush_1)
    # Specify the layout of the stroke data, in this case the stroke will have variable X, Y and Size properties.
    layout_mask: int = LayoutMask.X.value | LayoutMask.Y.value | LayoutMask.SIZE.value
    # Provide the stroke data - in this case 4 data points, each consisting of X, Y, Size
    path: list = [10., 10.7, 1.0, 21.0, 20.2, 2.0, 30.0, 12.4, 2.1, 40.0, 15.0, 1.5]
    # Create a spline object from the path data
    spline: Spline = Spline(layout_mask, path)
    # Create some style
    style: Style = Style(brush_uri=vector_brush_0.name)
    style.path_point_properties.red = 1.0
    style.path_point_properties.green = 0.0
    style.path_point_properties.blue = 0.0
    style.path_point_properties.alpha = 1.0
    # Create a stroke object. Note that it just exists, but is not in the model yet.
    stroke_0: Stroke = Stroke(sid=UUIDIdentifier.id_generator(), spline=spline, style=style)
    # Create a spline object - 9 data points, each consisting of X, Y, Size, Red, Green, Blue, Alpha
    spline_1: Spline = Spline(
        LayoutMask.X.value | LayoutMask.Y.value | LayoutMask.SIZE.value | LayoutMask.RED.value | LayoutMask.GREEN.value
        | LayoutMask.BLUE.value | LayoutMask.ALPHA.value,
        [10.0, 10.7, 1.0, 0.5, 0.0, 0.1, 1.0,
         21.0, 20.2, 2.0, 0.9, 0.4, 0.2, 0.8,
         30.0, 12.4, 2.1, 0.7, 0.1, 0.1, 0.7,
         40.0, 15.0, 1.5, 0.3, 0.5, 0.4, 1.0,
         50.0, 45.0, 1.0, 0.3, 0.5, 0.4, 1.0,
         41.0, 53.0, 1.1, 0.2, 0.3, 0.5, 0.9,
         33.0, 73.0, 1.2, 0.6, 0.7, 0.4, 0.8,
         20.0, 84.0, 1.3, 0.7, 0.8, 0.3, 0.7,
         10.0, 91.0, 1.1, 0.7, 0.9, 0.2, 0.6])
    spline_2: Spline = Spline(
        LayoutMask.X.value | LayoutMask.Y.value | LayoutMask.SIZE.value,
        [20., 20., 1.0, 21.0, 20.2, 2.0, 30.0, 12.4, 2.1, 40.0, 15.0, 1.5])
    # Create a style
    style_1: Style = Style(brush_uri=raster_brush_0.name)
    style_1.path_point_properties.rotation = 0.35
    # Create another style
    style_2: Style = Style(properties=path_point_properties,
                           brush_uri=vector_brush_1.name)

    # The render mode URI can also be app specific like app://blabla
    # The URI will://rasterization/3.0/blend-mode/SourceOver is assumed and must not be set.
    style_1.render_mode_uri = "will://rasterization/3.0/blend-mode/DestinationOver"
    # Create a stroke object. Note that it just exists, but is not in the model yet.
    stroke_1: Stroke = Stroke(UUIDIdentifier.id_generator(), spline=spline_1, style=style_1)
    stroke_2: Stroke = Stroke(UUIDIdentifier.id_generator(), spline=spline_2, style=style_2)
    # First you need a root group to contain the strokes
    root: StrokeGroupNode = StrokeGroupNode(UUIDIdentifier.id_generator())
    main_tree: InkTree = InkTree()
    ink_model.add_tree(main_tree)
    # Assign the group as the root of the main ink tree
    ink_model.ink_tree.root = root
    # Add a node for stroke 0
    stroke_node_0: StrokeNode = StrokeNode(stroke_0, StrokeFragment(0, 1, 0.0, 1.0))
    root.add(stroke_node_0)
    ink_model.knowledge_graph.append(SemanticTriple(stroke_node_0.uri, schema.IS, "will:ml/0.2/MathStroke"))
    # Add a node for stroke 1
    root.add(StrokeNode(stroke_1, StrokeFragment(0, 1, 0.0, 1.0)))
    # Add third node
    root.add(StrokeNode(stroke_2))

    # Adding view for handwriting recognition results
    hwr_tree: ViewTree = ViewTree(CommonViews.HWR_VIEW.value)
    hwr_root: StrokeGroupNode = StrokeGroupNode(UUIDIdentifier.id_generator())
    hwr_tree.root = hwr_root
    # Here you can add the same strokes as in the main tree, but you can organize them in a different way
    # (put them in different groups)
    # You are not supposed to add strokes that are not already in the main tree.
    hwr_root.add(StrokeNode(stroke_0, StrokeFragment(0, 1, 0.0, 1.0)))
    # A view node can refer to a fragment of a stroke.
    hwr_root.add(StrokeNode(stroke_1, StrokeFragment(0, 1, 0.0, 1.0)))
    node: InkNode = hwr_root.children[0]
    hwr_root.remove(node)
    hwr_tree.unregister_node(node)
    hwr_root.add(node)
    ink_model.add_view(hwr_tree)
    # The hwr root denotes a word
    ink_model.knowledge_graph.append(SemanticTriple(hwr_root.uri, schema.IS, schema.SegmentationSchema.WORD))
    ink_model.knowledge_graph.append(SemanticTriple(hwr_root.uri, schema.SegmentationSchema.HAS_CONTENT, "Wacom"))
    # We need a URI builder
    uri_builder: URIBuilder = URIBuilder()
    # Create a named entity
    named_entity_uri = uri_builder.build_named_entity_uri(UUIDIdentifier.id_generator())
    ink_model.knowledge_graph.append(SemanticTriple(hwr_root.uri, schema.NamedEntityRecognitionSchema.HAS_NAMED_ENTITY,
                                                    named_entity_uri))
    # Add knowledge for the named entity
    ink_model.knowledge_graph.append(SemanticTriple(named_entity_uri, "entityType", "Organization"))
    ink_model.knowledge_graph.append(SemanticTriple(named_entity_uri, "basedIn", "Japan"))
    ink_model.knowledge_graph.append(SemanticTriple(named_entity_uri, "hasConfidence", "0.85"))
    return ink_model


def pen_sensor_context(ink_model: InkModel, provider: InkInputProvider) -> Tuple[SensorChannelsContext,
                                                                                 List[SensorChannel]]:
    """
    Create a sensor context for a pen.

    Parameters
    ----------
    ink_model: InkModel
        The ink model.
    provider: InkInputProvider
        The input provider.

    Returns
    -------
    Tuple[SensorChannelsContext, List[SensorChannel]]
        Sensor channels context and the list of sensor channels.
    """
    input_device_pen: InputDevice = InputDevice()
    input_device_pen.properties.append(("dev.id", "345456567"))
    input_device_pen.properties.append(("dev.manufacturer", "Wacom"))
    ink_model.input_configuration.devices.append(input_device_pen)
    # Sensor channels
    sensor_channels_bluetooth: list = [
        SensorChannel(channel_type=InkSensorType.X, metric=InkSensorMetricType.LENGTH, resolution=1.0),
        SensorChannel(channel_type=InkSensorType.Y, metric=InkSensorMetricType.LENGTH, resolution=1.0),
        SensorChannel(channel_type=InkSensorType.TIMESTAMP, metric=InkSensorMetricType.TIME, resolution=1.0),
        SensorChannel(channel_type=InkSensorType.PRESSURE, metric=InkSensorMetricType.FORCE, resolution=1.0)
    ]
    # Create a sensor channel context
    scc_bluetooth: SensorChannelsContext = SensorChannelsContext(input_device_id=input_device_pen.id,
                                                                 ink_input_provider_id=provider.id,
                                                                 channels=sensor_channels_bluetooth)
    assert input_device_pen != provider
    return scc_bluetooth, sensor_channels_bluetooth


def test_spline():
    """
    Test spline structure.
    """
    spline = Spline(
        layout_mask=LayoutMask.X.value | LayoutMask.Y.value | LayoutMask.SIZE.value,
        data=[],
        ts=2.,
        tf=0.
    )
    assert spline.layout_mask == LayoutMask.X.value | LayoutMask.Y.value | LayoutMask.SIZE.value
    assert spline.ts == 2.
    assert spline.tf == 0.
    assert spline.data == []
    assert str(spline) == "<Spline : [mask:=11]>"


def test_identifier():
    """
    Test UUID identifier.
    """
    with pytest.raises(TypeError):
        InputDevice(identifier="Not valid")
    pen_device = InputDevice(UUIDIdentifier.id_generator())
    assert pen_device.method == IdentifiableMethod.MD5
    assert Identifier.uimid_to_h_form(None) == ""
    with pytest.raises(InkModelException):
        Identifier.str_to_uimid("Not valid")
    with pytest.raises(InkModelException):
        Identifier.from_bytes(b"12345678-1234-5678-1234-567812345678")
    uuid_identifier = UUIDIdentifier(UUIDIdentifier.id_generator())
    assert uuid_identifier.method == IdentifiableMethod.UUID
    assert str(uuid_identifier)


def test_path_point_properties():
    """
    Test path point properties structure.
    """
    path_point_properties = PathPointProperties(
        size=1.0,
        red=0.5,
        green=0.5,
        blue=0.5,
        alpha=1.0,
        rotation=0.0,
        scale_x=1.0,
        scale_y=1.0,
        scale_z=1.0,
        offset_x=0.0,
        offset_y=0.0,
        offset_z=0.0
    )
    path_point_properties_2 = PathPointProperties(
        size=1.0,
        red=0.5,
        green=0.5,
        blue=0.5,
        alpha=1.0,
        rotation=0.0,
        scale_x=1.0,
        scale_y=1.0,
        scale_z=1.0,
        offset_x=0.0,
        offset_y=0.0,
        offset_z=0.0
    )
    assert path_point_properties == path_point_properties_2
    assert path_point_properties != 1.0
    path_point_properties_2.size = 2.0
    assert path_point_properties != path_point_properties_2
    path_point_properties_2.size = 1.0
    path_point_properties_2.red = 0.7
    assert path_point_properties != path_point_properties_2
    path_point_properties_2.red = 0.5
    path_point_properties_2.green = 0.7
    assert path_point_properties != path_point_properties_2
    path_point_properties_2.green = 0.5
    path_point_properties_2.blue = 0.7
    assert path_point_properties != path_point_properties_2
    path_point_properties_2.blue = 0.5
    path_point_properties_2.alpha = 0.7
    assert path_point_properties != path_point_properties_2
    path_point_properties_2.alpha = 1.0
    path_point_properties_2.rotation = 0.7
    assert path_point_properties != path_point_properties_2
    path_point_properties_2.rotation = 0.0
    path_point_properties_2.scale_x = 2.0
    assert path_point_properties != path_point_properties_2
    path_point_properties_2.scale_x = 1.0
    path_point_properties_2.scale_y = 2.0
    assert path_point_properties != path_point_properties_2
    path_point_properties_2.scale_y = 1.0
    path_point_properties_2.scale_z = 2.0
    assert path_point_properties != path_point_properties_2
    path_point_properties_2.scale_z = 1.0
    path_point_properties_2.offset_x = 2.0
    assert path_point_properties != path_point_properties_2
    path_point_properties_2.offset_x = 0.0
    path_point_properties_2.offset_y = 2.0
    assert path_point_properties != path_point_properties_2
    path_point_properties_2.offset_y = 0.0
    path_point_properties_2.offset_z = 2.0
    assert path_point_properties != path_point_properties_2
    path_point_properties_2.offset_z = 0.0


def test_style():
    """
    Test style structure.
    """
    path_point_properties = PathPointProperties(
        size=1.0,
        red=0.5,
        green=0.5,
        blue=0.5,
        alpha=1.0,
        rotation=0.0,
        scale_x=1.0,
        scale_y=1.0,
        scale_z=1.0,
        offset_x=0.0,
        offset_y=0.0,
        offset_z=0.0
    )
    # Test the properties
    assert path_point_properties.size == 1.0
    assert path_point_properties.red == 0.5
    assert path_point_properties.green == 0.5
    assert path_point_properties.blue == 0.5
    assert path_point_properties.alpha == 1.0
    assert path_point_properties.rotation == 0.0
    assert path_point_properties.scale_x == 1.0
    assert path_point_properties.scale_y == 1.0
    assert path_point_properties.scale_z == 1.0
    assert path_point_properties.offset_x == 0.0
    assert path_point_properties.offset_y == 0.0
    assert path_point_properties.offset_z == 0.0

    style = Style(
        properties=path_point_properties,
        brush_uri="app://qa-test-app/vector-brush/MyTriangleBrush",
        particles_random_seed=0,
        render_mode_uri=BlendModeURIs.SOURCE_OVER
    )
    # Test the properties
    assert style.path_point_properties.size == 1.0
    assert style.path_point_properties.red == 0.5
    assert style.path_point_properties.green == 0.5
    assert style.path_point_properties.blue == 0.5
    assert style.path_point_properties.alpha == 1.0
    assert style.path_point_properties.rotation == 0.0
    assert style.path_point_properties.scale_x == 1.0
    assert style.path_point_properties.scale_y == 1.0
    assert style.path_point_properties.scale_z == 1.0
    assert style.path_point_properties.offset_x == 0.0
    assert style.path_point_properties.offset_y == 0.0
    assert style.path_point_properties.offset_z == 0.0
    assert style.brush_uri == "app://qa-test-app/vector-brush/MyTriangleBrush"
    assert style.particles_random_seed == 0
    assert style.render_mode_uri == BlendModeURIs.SOURCE_OVER
    assert str(style) == ('<Style : [id:=app://qa-test-app/vector-brush/MyTriangleBrush, '
                          'particles_random_seed:=0, render mode:=will://rasterization/3.0/blend-mode/SourceOver>')
    path_point_properties_2 = PathPointProperties(
        size=2.0,
        red=0.5,
        green=0.5,
        blue=0.5,
        alpha=1.0,
        rotation=0.0,
        scale_x=1.0,
        scale_y=1.0,
        scale_z=1.0,
        offset_x=0.0,
        offset_y=0.0,
        offset_z=0.0
    )
    style_other = Style(
        properties=path_point_properties_2,
        brush_uri="app://qa-test-app/vector-brush/MyTriangleBrush",
        particles_random_seed=0,
        render_mode_uri=BlendModeURIs.SOURCE_OVER
    )
    assert style != style_other
    style_other = Style(
        properties=path_point_properties,
        brush_uri="app://qa-test-app/vector-brush/MyTriangleBrush2",
        particles_random_seed=0,
        render_mode_uri=BlendModeURIs.SOURCE_OVER
    )
    assert style != style_other
    style_other = Style(
        properties=path_point_properties,
        brush_uri="app://qa-test-app/vector-brush/MyTriangleBrush",
        particles_random_seed=3,
        render_mode_uri=BlendModeURIs.SOURCE_OVER
    )
    assert style != style_other
    style_other = Style(
        properties=path_point_properties,
        brush_uri="app://qa-test-app/vector-brush/MyTriangleBrush",
        particles_random_seed=0,
        render_mode_uri=BlendModeURIs.COPY
    )
    assert style != style_other
    assert style != path_point_properties


def test_knowledge_graph():
    """Creates an ink model from the scratch."""
    # Create the model
    ink_model: InkModel = create_ink_model()
    hwr_root: StrokeGroupNode = ink_model.view_root(name=CommonViews.HWR_VIEW.value)
    # Check statement filters
    assert str(ink_model.knowledge_graph)
    assert len(ink_model.knowledge_graph.all_statements_for(subject=hwr_root.uri)) == 3
    assert len(ink_model.knowledge_graph.filter(obj="Wacom")) == 1
    assert len(ink_model.knowledge_graph.filter(obj="Not found")) == 0
    assert len(ink_model.knowledge_graph.filter(predicate=schema.IS)) == 2
    for stmt in ink_model.knowledge_graph:
        assert str(stmt)
        assert stmt.subject is not None
        assert stmt.predicate is not None
        assert stmt.object is not None

    triples: List[SemanticTriple] = ink_model.knowledge_graph.filter(predicate=schema.SegmentationSchema.HAS_CONTENT,
                                                                     obj="Wacom")
    assert len(triples) == 1
    triple: SemanticTriple = triples[0]
    ink_model.remove_semantic_triple(triple.subject, triple.predicate, triple.object)
    assert len(ink_model.knowledge_graph.filter(predicate=schema.SegmentationSchema.HAS_CONTENT, obj="Wacom")) == 0
    assert (ink_model.knowledge_graph.determine_sem_type(hwr_root, typedef_pred=schema.IS)
            == schema.SegmentationSchema.WORD)
    ink_model.clear_knowledge_graph()
    assert len(ink_model.knowledge_graph.statements) == 0
    assert ink_model.knowledge_graph.determine_sem_type(hwr_root) is None


def test_semantic_triple():
    """
    Test the semantic triple structure.
    """
    uri: str = uuid.uuid4().hex
    triple: SemanticTriple = SemanticTriple(subject=uri, predicate=schema.IS,
                                            obj=schema.SegmentationSchema.WORD)
    triple_2: SemanticTriple = SemanticTriple(subject=uri, predicate=schema.IS,
                                              obj=schema.SegmentationSchema.WORD)
    assert triple == triple_2
    triple_2.predicate = schema.SegmentationSchema.HAS_CONTENT
    triple_2.object = "Wacom"
    assert triple != triple_2
    assert triple != "Different type"
    assert triple.__json__() == {
        "subject": uri,
        "predicate": schema.IS,
        "object": schema.SegmentationSchema.WORD
    }


def test_empty_ink_model():
    """
    Test the empty ink model.
    """
    ink_model: InkModel = InkModel(SupportedFormats.UIM_VERSION_3_1_0.value)
    assert not ink_model.has_ink_data()
    assert not ink_model.has_ink_structure()
    assert not ink_model.has_brushes()
    assert not ink_model.has_knowledge_graph()
    assert not ink_model.has_properties()
    assert len(ink_model.views) == 0
    assert ink_model != InkModel(SupportedFormats.UIM_VERSION_3_0_0.value)
    assert ink_model != SupportedFormats.UIM_VERSION_3_1_0
    ink_model.transform = [[1., 0., 0., 0.], [0., 1., 0., 0.], [0., 0., 1., 0.], [0., 0., 0., 1.]]
    ink_model_2: InkModel = InkModel(SupportedFormats.UIM_VERSION_3_1_0.value)
    ink_model_2.transform = [[2., 0., 0., 0.], [0., 2., 0., 0.], [0., 0., 2., 0.], [0., 0., 0., 1.]]
    assert ink_model != ink_model_2
    assert SupportedFormats.UIM_VERSION_3_1_0.value != ink_model
    assert str(ink_model)


def test_ink_tree():
    """
    Test the empty ink model.
    """
    ink_model: InkModel = InkModel(SupportedFormats.UIM_VERSION_3_1_0.value)
    assert ink_model.ink_tree is None
    ink_tree: InkTree = InkTree()
    assert str(ink_tree)
    with pytest.raises(InkModelException):
        ink_tree.root = None
    with pytest.raises(ValueError):
        PreOrderEnumerator(ink_tree.root)
    ink_model.add_tree(ink_tree)
    assert ink_model.ink_tree == ink_tree
    assert len(ink_model.views) == 0
    root_id: uuid.UUID = UUIDIdentifier.id_generator()
    ink_tree_2: InkTree = InkTree()
    ink_tree_2.root = StrokeGroupNode(root_id)
    root_node: StrokeGroupNode = StrokeGroupNode(UUIDIdentifier.id_generator())
    ink_tree_2.root = root_node
    ink_tree_2.root.add(StrokeNode(Stroke(UUIDIdentifier.id_generator())))
    assert ink_tree != ink_tree_2
    ink_tree.root = StrokeGroupNode(root_id)
    ink_tree.root.add(StrokeNode(Stroke(UUIDIdentifier.id_generator())))
    assert ink_tree != ink_tree_2
    ink_tree_2.unregister_node(root_node)


def test_bounding_boxes():
    """Test to calculate bounding boxes."""
    # Create the model
    ink_model: InkModel = create_ink_model()
    hwr_root: StrokeGroupNode = ink_model.view_root(name=CommonViews.HWR_VIEW.value)
    assert hwr_root.group_bounding_box == BoundingBox(0., 0, 0, 0)
    for stroke in ink_model.ink_tree:
        assert stroke.group_bounding_box == BoundingBox(0., 0, 0, 0)

    ink_model.calculate_bounds_recursively(hwr_root)
    ink_model.calculate_bounds_recursively(ink_model.ink_tree.root)
    assert hwr_root.group_bounding_box == BoundingBox(x=10., y=10.7, width=40., height=80.3)
    enclosing = hwr_root.group_bounding_box.enclosing_bounding_box(BoundingBox(0., 0., 10., 10.))
    # The enclosing bounding box from 0., 0. and the width is 50, as the width of the stroke is 40 and the x is 10
    # so the rightmost point is 50. The height is 91.0, as the height of the stroke is 80.3 and the y is 10.7 so the
    # topmost point is 91.0
    assert enclosing == BoundingBox(0., 0., 50., 91.0)
    assert enclosing != ink_model
    for stroke in ink_model.ink_tree:
        bbox: BoundingBox = stroke.group_bounding_box
        assert bbox.x > 0.
        assert bbox.y > 0.
        assert bbox.width > 0.
        assert bbox.height > 0.
    for stroke in ink_model.view(CommonViews.HWR_VIEW.value):
        bbox: BoundingBox = stroke.group_bounding_box
        assert bbox.x > 0.
        assert bbox.y > 0.
        assert bbox.width > 0.
        assert bbox.height > 0.


def test_sensor_data():
    """
    Test the sensor data structures.
    """
    ink_model: InkModel = create_ink_model(with_sensor_data=False)
    ink_model_2: InkModel = create_ink_model(with_sensor_data=False)
    input_context = ink_model_2.input_configuration.input_contexts[0]

    # Adding additional sensor data
    sensor_data_id: uuid.UUID = UUIDIdentifier.id_generator()
    sensor_data_1: SensorData = SensorData(sensor_data_id, input_context_id=input_context.id,  state=InkState.PLANE)
    sensor_channels_tablet = ink_model.input_configuration.sensor_contexts[0].sensor_channels_contexts[0].channels
    sensor_data_1.add_timestamp_data(sensor_channels_tablet[2], [0., 0.1, 0.2, 0.3, 0.4])
    sensor_data_1.add_data(sensor_channels_tablet[0], [100.4, 103.7, 110.1])
    sensor_data_1.add_data(sensor_channels_tablet[1], [200.1, 202.0, 207.0])
    sensor_data_1.add_data(sensor_channels_tablet[3], [0.1, 0.4, 0.2])
    sensor_data_1.add_data(sensor_channels_tablet[4], [40., 51.0, 52.0])
    sensor_data_1.add_data(sensor_channels_tablet[5], [40., 51.0, 52.0])
    ink_model.sensor_data.add(sensor_data_1)
    sensor_data_2: SensorData = SensorData(sensor_data_id, input_context_id=input_context.id,
                                           state=InkState.PLANE)
    sensor_channels_tablet = ink_model_2.input_configuration.sensor_contexts[0].sensor_channels_contexts[0].channels
    sensor_data_2.add_timestamp_data(sensor_channels_tablet[2], [1., 0.1, 0.2, 0.3, 0.4])
    sensor_data_2.add_data(sensor_channels_tablet[0], [200.4, 103.7, 110.1])
    sensor_data_2.add_data(sensor_channels_tablet[1], [300.1, 202.0, 207.0])
    sensor_data_2.add_data(sensor_channels_tablet[3], [4.1, 0.4, 0.2])
    sensor_data_2.add_data(sensor_channels_tablet[4], [60., 51.0, 52.0])
    sensor_data_2.add_data(sensor_channels_tablet[5], [60., 51.0, 52.0])
    ink_model_2.sensor_data.add(sensor_data_2)
    assert ink_model.sensor_data != ink_model_2.sensor_data
    assert sensor_data_1 != sensor_data_2
    sensor_data_3: SensorData = SensorData(UUIDIdentifier.id_generator(), input_context_id=input_context.id,
                                           state=InkState.PLANE)
    sensor_channels_tablet = ink_model_2.input_configuration.sensor_contexts[0].sensor_channels_contexts[0].channels
    sensor_data_3.add_timestamp_data(sensor_channels_tablet[2], [1., 0.1, 0.2, 0.3, 0.4])
    sensor_data_3.add_data(sensor_channels_tablet[0], [200.4, 103.7, 110.1])
    sensor_data_3.add_data(sensor_channels_tablet[1], [300.1, 202.0, 207.0])
    sensor_data_3.add_data(sensor_channels_tablet[3], [4.1, 0.4, 0.2])
    sensor_data_3.add_data(sensor_channels_tablet[4], [60., 51.0, 52.0])
    sensor_data_3.add_data(sensor_channels_tablet[5], [60., 51.0, 52.0])
    assert sensor_data_2 != sensor_data_3


def test_brushes():
    """
    Test the brush structures.
    """
    ink_model: InkModel = InkModel(SupportedFormats.UIM_VERSION_3_1_0.value)
    points_1: list = [(10, 20), (0, 20), (0, 0)]
    points_2: list = [(1, 1), (0, 1), (0, 0)]
    bp_1: BrushPolygon = BrushPolygon(min_scale=0., points=points_1)
    bp_2: BrushPolygon = BrushPolygon(min_scale=0., points=points_2)
    bp_3: BrushPolygon = BrushPolygon(min_scale=1., points=points_1, indices=[0, 1, 2])
    bp_4: BrushPolygon = BrushPolygon(min_scale=1., points=points_1, indices=[0, 1, 3])
    brush_polygons: list = [bp_1, bp_2]
    # Check all variants of __eq__
    assert bp_1 != bp_2
    assert bp_1 != ink_model
    assert bp_1 != bp_3
    assert bp_3 != bp_4
    assert str(bp_1)
    assert bp_1 == bp_1
    assert bp_2 == bp_2
    assert bp_1.min_scale == 0.
    assert bp_1.points == points_1
    assert bp_2.min_scale == 0.
    assert bp_2.points == points_2
    assert bp_1.coord_x == [10, 0, 0]
    assert bp_1.coord_y == [20, 20, 0]
    assert bp_2.coord_x == [1, 0, 0]
    assert bp_2.coord_y == [1, 1, 0]

    b = BrushPolygon(min_scale=0., points=[])
    assert b.coord_x == []

    b = BrushPolygon(min_scale=0., points=[])
    assert b.coord_y == []
    assert b.coord_z == []
    assert b.indices == []

    with pytest.raises(InkModelException):
        b = BrushPolygon(min_scale=0., points=[1, 2, 3])
        assert b.coord_x

    with pytest.raises(InkModelException):
        b = BrushPolygon(min_scale=0., points=[1, 2, 3])
        assert b.coord_y

    points_3d: list = [(10, 20, 30), (0, 20, 30), (0, 0, 30)]
    bp_3: BrushPolygon = BrushPolygon(min_scale=0., points=points_3d)
    assert bp_3.coord_x == [10, 0, 0]
    assert bp_3.coord_y == [20, 20, 0]
    assert bp_3.coord_z == [30, 30, 30]

    # Create the brush object using polygons
    vector_brush_0: VectorBrush = VectorBrush(
        name="app://qa-test-app/vector-brush/MyTriangleBrush",
        prototypes=brush_polygons,
        spacing=4
    )
    assert str(vector_brush_0)
    assert vector_brush_0.name == "app://qa-test-app/vector-brush/MyTriangleBrush"
    assert vector_brush_0.spacing == 4
    assert vector_brush_0.prototypes == brush_polygons
    # Add a brush specified with shape Uris
    bpu_1: BrushPolygonUri = BrushPolygonUri("will://brush/3.0/shape/Circle?precision=20&radius=1", min_scale=0.)
    bpu_2: BrushPolygonUri = BrushPolygonUri("will://brush/3.0/shape/Circle?precision=20&radius=0.5", min_scale=4.)

    poly_uris: list = [
        bpu_1, bpu_2
    ]
    assert bpu_1 != bpu_2
    assert bpu_1 != ink_model
    assert bpu_1 == bpu_1
    assert bpu_2 == bpu_2
    assert str(bpu_1)
    vector_brush_1: VectorBrush = VectorBrush(
        "app://qa-test-app/vector-brush/MyEllipticBrush",
        poly_uris)
    raster_brush_0: RasterBrush = RasterBrush(
        name="app://qa-test-app/raster-brush/MyRasterBrush",
        spacing=10., scattering=5., rotation=RotationMode.TRAJECTORY, shape_textures=[bytes([10, 20]),
                                                                                      bytes([30, 20])],
        fill_width=2.0, fill_height=0.3,
        fill_texture=bytes([10, 10, 20, 15, 17, 20, 25, 16, 34, 255, 23, 0, 34, 255, 23, 255]),
        randomize_fill=False, blend_mode=BlendMode.SOURCE_OVER)
    assert vector_brush_0 != raster_brush_0
    assert raster_brush_0 != vector_brush_0
    # Add it to the model
    raster_brush_1: RasterBrush = RasterBrush(
        name="app://qa-test-app/raster-brush/MyRasterBrush1",
        spacing=10.0, scattering=5., rotation=RotationMode.TRAJECTORY, fill_width=2., fill_height=0.3,
        fill_texture_uri="app://qa-test-app/raster-brush-fill/mixedShapesGL",
        shape_texture_uris=[
            "app://qa-test-app/raster-brush-shape/mixedShapesGL_128x128",
            "app://qa-test-app/raster-brush-shape/mixedShapesGL_64x64",
            "app://qa-test-app/raster-brush-shape/mixedShapesGL_32x32",
            "app://qa-test-app/raster-brush-shape/mixedShapesGL_16x16"
        ], randomize_fill=False, blend_mode=BlendMode.SOURCE_OVER)
    raster_brush_2: RasterBrush = RasterBrush(
        name="app://qa-test-app/raster-brush/MyRasterBrush1",
        spacing=10.0, scattering=5., rotation=RotationMode.TRAJECTORY, fill_width=2.5, fill_height=0.3,
        fill_texture_uri="app://qa-test-app/raster-brush-fill/mixedShapesGL",
        shape_texture_uris=[
            "app://qa-test-app/raster-brush-shape/mixedShapesGL_128x128",
            "app://qa-test-app/raster-brush-shape/mixedShapesGL_64x64",
            "app://qa-test-app/raster-brush-shape/mixedShapesGL_32x32"
        ], randomize_fill=False, blend_mode=BlendMode.SOURCE_OVER
    )
    raster_brush_3: RasterBrush = RasterBrush(
        name="app://qa-test-app/raster-brush/MyRasterBrush1",
        spacing=10.0, scattering=5., rotation=RotationMode.TRAJECTORY, fill_width=2.5, fill_height=0.3,
        fill_texture_uri="app://qa-test-app/raster-brush-fill/mixedShapesGL",
        shape_texture_uris=[
            "app://qa-test-app/raster-brush-shape/mixedShapesGL_128x128",
            "app://qa-test-app/raster-brush-shape/mixedShapesGL_64x64",
            "app://qa-test-app/raster-brush-shape/mixedShapesGL_32x32",
            "app://qa-test-app/raster-brush-shape/mixedShapesGL_16x16"
        ], randomize_fill=False, blend_mode=BlendMode.COPY
    )
    raster_brush_4: RasterBrush = RasterBrush(
        name="app://qa-test-app/raster-brush/MyRasterBrush1",
        spacing=10.0, scattering=5., rotation=RotationMode.TRAJECTORY, fill_width=2.5, fill_height=0.3,
        fill_texture_uri="app://qa-test-app/raster-brush-fill/mixedShapesGL",
        shape_texture_uris=[
            "app://qa-test-app/raster-brush-shape/mixedShapesGL_128x128",
            "app://qa-test-app/raster-brush-shape/mixedShapesGL_64x64",
            "app://qa-test-app/raster-brush-shape/mixedShapesGL_32x32",
            "app://qa-test-app/raster-brush-shape/mixedShapesGL_16x16"
        ], randomize_fill=True, blend_mode=BlendMode.COPY
    )
    raster_brush_5: RasterBrush = RasterBrush(
        name="app://qa-test-app/raster-brush/MyRasterBrush1",
        spacing=10.0, scattering=5., rotation=RotationMode.TRAJECTORY, fill_width=2.6, fill_height=0.3,
        fill_texture_uri="app://qa-test-app/raster-brush-fill/mixedShapesGL",
        shape_texture_uris=[
            "app://qa-test-app/raster-brush-shape/mixedShapesGL_128x128",
            "app://qa-test-app/raster-brush-shape/mixedShapesGL_64x64",
            "app://qa-test-app/raster-brush-shape/mixedShapesGL_32x32",
            "app://qa-test-app/raster-brush-shape/mixedShapesGL_16x16"
        ], randomize_fill=True, blend_mode=BlendMode.COPY
    )
    raster_brush_6: RasterBrush = RasterBrush(
        name="app://qa-test-app/raster-brush/MyRasterBrush1",
        spacing=10.0, scattering=5., rotation=RotationMode.TRAJECTORY, fill_width=2.6, fill_height=0.7,
        fill_texture_uri="app://qa-test-app/raster-brush-fill/mixedShapesGL",
        shape_texture_uris=[
            "app://qa-test-app/raster-brush-shape/mixedShapesGL_128x128",
            "app://qa-test-app/raster-brush-shape/mixedShapesGL_64x64",
            "app://qa-test-app/raster-brush-shape/mixedShapesGL_32x32",
            "app://qa-test-app/raster-brush-shape/mixedShapesGL_16x16"
        ], randomize_fill=True, blend_mode=BlendMode.COPY
    )
    raster_brush_7: RasterBrush = RasterBrush(
        name="app://qa-test-app/raster-brush/MyRasterBrush1",
        spacing=10.0, scattering=5., rotation=RotationMode.RANDOM, fill_width=2.6, fill_height=0.7,
        fill_texture_uri="app://qa-test-app/raster-brush-fill/mixedShapesGL",
        shape_texture_uris=[
            "app://qa-test-app/raster-brush-shape/mixedShapesGL_128x128",
            "app://qa-test-app/raster-brush-shape/mixedShapesGL_64x64",
            "app://qa-test-app/raster-brush-shape/mixedShapesGL_32x32",
            "app://qa-test-app/raster-brush-shape/mixedShapesGL_16x16"
        ], randomize_fill=True, blend_mode=BlendMode.COPY
    )
    raster_brush_8: RasterBrush = RasterBrush(
        name="app://qa-test-app/raster-brush/MyRasterBrush1",
        spacing=10.0, scattering=5., rotation=RotationMode.RANDOM, fill_width=2.6, fill_height=0.7,
        fill_texture_uri="app://qa-test-app/raster-brush-fill/mixedShapesGL",
        shape_texture_uris=[
            "app://qa-test-app/raster-brush-shape/mixedShapesGL_128x128",
            "app://qa-test-app/raster-brush-shape/mixedShapesGL_64x64",
            "app://qa-test-app/raster-brush-shape/mixedShapesGL_32x32",
            "app://qa-test-app/raster-brush-shape/mixedShapesGL_16x16"
        ], randomize_fill=True, blend_mode=BlendMode.SOURCE_OVER
    )
    raster_brush_9: RasterBrush = RasterBrush(
        name="app://qa-test-app/raster-brush/MyRasterBrush1",
        spacing=10.0, scattering=5., rotation=RotationMode.RANDOM, fill_width=2.6, fill_height=0.7,
        fill_texture=bytes([10, 10, 20, 15, 17, 20, 25, 16, 34, 255, 23, 0, 34, 255, 23, 255]),
        fill_texture_uri="app://qa-test-app/raster-brush-fill/mixedShapesGL",
        shape_texture_uris=[
            "app://qa-test-app/raster-brush-shape/mixedShapesGL_128x128",
            "app://qa-test-app/raster-brush-shape/mixedShapesGL_64x64",
            "app://qa-test-app/raster-brush-shape/mixedShapesGL_32x32",
            "app://qa-test-app/raster-brush-shape/mixedShapesGL_16x16"
        ], randomize_fill=True, blend_mode=BlendMode.SOURCE_OVER
    )
    raster_brush_10: RasterBrush = RasterBrush(
        name="app://qa-test-app/raster-brush/MyRasterBrush1",
        spacing=10.0, scattering=5., rotation=RotationMode.RANDOM, fill_width=2.6, fill_height=0.7,
        fill_texture=bytes([10, 10, 20, 15, 17, 20, 25, 16, 34, 255, 23, 0, 34, 255, 23, 255]),
        fill_texture_uri="app://qa-test-app/raster-brush-fill/mixedShapesGLDiff",
        shape_texture_uris=[
            "app://qa-test-app/raster-brush-shape/mixedShapesGL_128x128",
            "app://qa-test-app/raster-brush-shape/mixedShapesGL_64x64",
            "app://qa-test-app/raster-brush-shape/mixedShapesGL_32x32",
            "app://qa-test-app/raster-brush-shape/mixedShapesGL_16x16"
        ], randomize_fill=True, blend_mode=BlendMode.SOURCE_OVER
    )
    raster_brush_11: RasterBrush = RasterBrush(
        name="app://qa-test-app/raster-brush/MyRasterBrush1",
        spacing=10.0, scattering=5., rotation=RotationMode.RANDOM, fill_width=2.6, fill_height=0.7,
        fill_texture=bytes([10, 10, 20, 15, 17, 20, 25, 16, 34, 255, 23, 0, 34, 255, 23, 255]),
        fill_texture_uri="app://qa-test-app/raster-brush-fill/mixedShapesGLDiff",
        shape_texture_uris=[
            "app://qa-test-app/raster-brush-shape/mixedShapesGL_128x128",
            "app://qa-test-app/raster-brush-shape/mixedShapesGL_64x64",
            "app://qa-test-app/raster-brush-shape/mixedShapesGL_32x32"
        ], randomize_fill=True, blend_mode=BlendMode.SOURCE_OVER
    )
    raster_brush_12: RasterBrush = RasterBrush(
        name="app://qa-test-app/raster-brush/MyRasterBrush1",
        spacing=14.0, scattering=5., rotation=RotationMode.RANDOM, fill_width=2.6, fill_height=0.7,
        fill_texture=bytes([10, 10, 20, 15, 17, 20, 25, 16, 34, 255, 23, 0, 34, 255, 23, 255]),
        fill_texture_uri="app://qa-test-app/raster-brush-fill/mixedShapesGLDiff",
        shape_texture_uris=[
            "app://qa-test-app/raster-brush-shape/mixedShapesGL_128x128",
            "app://qa-test-app/raster-brush-shape/mixedShapesGL_64x64",
            "app://qa-test-app/raster-brush-shape/mixedShapesGL_32x32"
        ], randomize_fill=True, blend_mode=BlendMode.SOURCE_OVER
    )
    raster_brush_13: RasterBrush = RasterBrush(
        name="app://qa-test-app/raster-brush/MyRasterBrush1",
        spacing=14.0, scattering=7., rotation=RotationMode.RANDOM, fill_width=2.6, fill_height=0.7,
        fill_texture=bytes([10, 10, 20, 15, 17, 20, 25, 16, 34, 255, 23, 0, 34, 255, 23, 255]),
        fill_texture_uri="app://qa-test-app/raster-brush-fill/mixedShapesGLDiff",
        shape_texture_uris=[
            "app://qa-test-app/raster-brush-shape/mixedShapesGL_128x128",
            "app://qa-test-app/raster-brush-shape/mixedShapesGL_64x64",
            "app://qa-test-app/raster-brush-shape/mixedShapesGL_32x32"
        ], randomize_fill=True, blend_mode=BlendMode.SOURCE_OVER
    )
    raster_brush_14: RasterBrush = RasterBrush(
        name="app://qa-test-app/raster-brush/MyRasterBrush1",
        spacing=14.0, scattering=7., rotation=RotationMode.RANDOM, fill_width=2.6, fill_height=0.7,
        fill_texture=bytes([10, 10, 20, 15, 17, 20, 25, 16, 34, 255, 23, 0, 34, 255, 23, 255]),
        fill_texture_uri="app://qa-test-app/raster-brush-fill/mixedShapesGLDiff",
        shape_textures=[
            bytes([10, 10, 20, 15, 17, 20, 25, 16, 34, 255, 23, 0, 34, 255, 23, 255])
        ],
        shape_texture_uris=[
            "app://qa-test-app/raster-brush-shape/mixedShapesGL_128x128",
            "app://qa-test-app/raster-brush-shape/mixedShapesGL_64x64",
            "app://qa-test-app/raster-brush-shape/mixedShapesGL_32x32"
        ], randomize_fill=True, blend_mode=BlendMode.SOURCE_OVER
    )
    # Check all variants of __eq__
    assert raster_brush_0 != raster_brush_1
    assert raster_brush_1 != raster_brush_2
    assert raster_brush_2 != raster_brush_3
    assert raster_brush_3 != raster_brush_4
    assert raster_brush_4 != raster_brush_5
    assert raster_brush_5 != raster_brush_6
    assert raster_brush_6 != raster_brush_7
    assert raster_brush_7 != raster_brush_8
    assert raster_brush_8 != raster_brush_9
    assert raster_brush_9 != raster_brush_10
    assert raster_brush_10 != raster_brush_11
    assert raster_brush_11 != raster_brush_12
    assert raster_brush_12 != raster_brush_13
    assert raster_brush_13 != raster_brush_14
    # Check add and remove
    ink_model.brushes.add_raster_brush(raster_brush_0)
    ink_model.brushes.add_raster_brush(raster_brush_1)
    ink_model.brushes.add_vector_brush(vector_brush_0)
    ink_model.brushes.add_vector_brush(vector_brush_1)
    assert len(ink_model.brushes.vector_brushes) == 2
    assert len(ink_model.brushes.raster_brushes) == 2
    assert ink_model.brushes.vector_brushes[0] == vector_brush_0
    assert ink_model.brushes.raster_brushes[0] == raster_brush_0
    ink_model.brushes.remove_vector_brush(vector_brush_1.name)
    ink_model.brushes.remove_raster_brush(raster_brush_2.name)
    assert len(ink_model.brushes.vector_brushes) == 1
    assert len(ink_model.brushes.raster_brushes) == 1
    assert str(ink_model.brushes)


def test_validate_views():
    """
    Test the view structure.
    """
    # Create the model
    ink_model: InkModel = create_ink_model()
    # Check the views
    test_tree: ViewTree = ViewTree("Test")
    test_tree_2: ViewTree = ViewTree("Test")
    assert str(test_tree)
    root_node: StrokeGroupNode = StrokeGroupNode(UUIDIdentifier.id_generator())
    test_tree.root = root_node
    test_tree_2.root = StrokeGroupNode(UUIDIdentifier.id_generator())
    assert test_tree != test_tree_2
    with pytest.raises(InkModelException):
        test_tree.root = None
    # Remove the tree
    root_node: StrokeGroupNode = StrokeGroupNode(UUIDIdentifier.id_generator())
    test_tree.root = root_node
    assert str(root_node)
    with pytest.raises(InkModelException):
        test_tree.root.add(root_node)
    assert test_tree.model is None
    test_tree.root.add(StrokeNode(Stroke(UUIDIdentifier.id_generator())))
    test_tree_2.root.add(StrokeNode(Stroke(UUIDIdentifier.id_generator())))
    assert test_tree != test_tree_2
    for ch in test_tree.root.children:
        assert ch.root == test_tree.root
    main_tree: InkTree = ink_model.ink_tree
    assert str(main_tree)
    hwr_tree: ViewTree = ink_model.view(CommonViews.HWR_VIEW.value)
    with pytest.raises(InkModelException):
        ink_model.add_view(hwr_tree)
    assert len(ink_model.views) == 1
    assert main_tree != ink_model
    assert main_tree != hwr_tree
    assert main_tree != test_tree
    assert str(test_tree)
    assert str(hwr_tree)
    # Remove the tree
    ink_model.remove_tree(hwr_tree.name)
    ink_model.remove_tree(ink_model.ink_tree.name)
    assert ink_model.ink_tree is None
    assert len(ink_model.views) == 0
    # Remove a non-existing tree
    ink_model.remove_tree("Not found")
    # Add the view to the model
    ink_model.add_view(hwr_tree)
    ink_model.add_tree(main_tree)
    with pytest.raises(InkModelException):
        ink_model.add_tree(InkTree())
    assert len(ink_model.views) == 1
    assert ink_model.ink_tree is not None
    assert hwr_tree.model == ink_model
    ink_model.clear_views()
    assert len(ink_model.views) == 0
    ink_model.add_view(hwr_tree)
    assert len(ink_model.views) == 1
    with pytest.raises(KeyError):
        ink_model.view("Not found")
    ink_model.remove_node(main_tree.root.children[0])


def test_validate_serialise_deserialize():
    """Creates an ink model from the scratch."""
    # Create the model
    ink_model: InkModel = create_ink_model()
    # Save the model, this will overwrite an existing file
    new_model_3_1_0: InkModel = UIMParser().parse(UIMEncoder310().encode(ink_model))
    assert ink_model == new_model_3_1_0


def test_uri_builder():
    """
    Test the URI builder.
    """
    builder: URIBuilder = URIBuilder()
    model_id: uuid.UUID = UUIDIdentifier.id_generator()
    gen_uri: str = builder.build_uri(sub_path='sub', model_id=model_id)
    assert gen_uri == f'uim:{model_id}/sub/'
    gen_uri = builder.build_uri(sub_path='sub')
    assert gen_uri == f'uim:sub/'
    with pytest.raises(InkModelException):
        builder.build_named_entity_uri(uimid='uim:sub/', model_id=model_id)
    named_entity_id: uuid.UUID = UUIDIdentifier.id_generator()
    gen_uri = builder.build_named_entity_uri(uimid=named_entity_id, model_id=model_id)
    assert gen_uri == f'uim:{model_id}/ne/{named_entity_id}'
    stroke_uuid: uuid.UUID = UUIDIdentifier.id_generator()
    stroke_uri: str = builder.build_stroke_uri(stroke_uuid)
    assert stroke_uri == f'uim:stroke/{stroke_uuid}'
    ink_node: InkNode = StrokeNode(Stroke(UUIDIdentifier.id_generator()))
    ink_node.transient_tag = "tag"
    assert ink_node.transient_tag == "tag"
    with pytest.raises(InkModelException):
        assert ink_node.root
    assert str(ink_node)
    with pytest.raises(InkModelException):
        builder.build_node_uri(ink_node, SupportedFormats.UIM_VERSION_3_1_0)
    tree_uri = builder.build_tree_uri("test")
    assert tree_uri == "uim:tree/test"
    with pytest.raises(InkModelException):
        builder.build_node_uri(model_id, SupportedFormats.UIM_VERSION_3_1_0)
    builder.build_entity_uri(uimid=UUIDIdentifier.id_generator(), model_id=model_id)
    with pytest.raises(InkModelException):
        builder.build_entity_uri(uimid="uim:sub/", model_id=model_id)


def test_model_strokes():
    """
    Test the strokes in the model.
    """
    ink_model: InkModel = create_ink_model()
    # Get all strokes
    hwr_strokes = ink_model.collect_strokes(ink_model.views[0])
    assert len(hwr_strokes) == 2
    main_strokes = ink_model.collect_strokes(ink_model.ink_tree)
    assert len(main_strokes) == 3



# -*- coding: utf-8 -*-
# Copyright © 2024-present Wacom Authors. All Rights Reserved.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
import uuid
from typing import List, Dict, Any

from uim.model.helpers.treeiterator import PreOrderEnumerator
from uim.model.ink import InkModel
from uim.model.semantics import schema
from uim.model.semantics.node import InkNode, StrokeGroupNode, StrokeNode


def __collected_stroke_ids__(node: StrokeGroupNode) -> List[uuid.UUID]:
    """
    Collecting all stroke ids from the node.
    Parameters
    ----------
    node: `StrokeGroupNode`
        Stroke group node

    Returns
    -------
    strokes: `List[uuid.UUID]`
        List of stroke ids
    """
    strokes: List[uuid.UUID] = []
    for child in node.children:
        if isinstance(child, StrokeNode):
            strokes.append(child.stroke.id)
        elif isinstance(child, StrokeGroupNode):
            strokes.extend(__collected_stroke_ids__(child))
    return strokes


def uim_schema_semantics_from(ink_model: InkModel, semantic_view: str = schema.CommonViews.HWR_VIEW.value)\
        -> List[Dict[str, Any]]:
    """
    Extract schema semantics from the ink model.
    Parameters
    ----------
    ink_model: `InkModel`
        Ink model
    semantic_view: `str`
        Semantic view

    Returns
    -------
    elements: `List[Dict[str, Any]]`
        List of schema semantics elements. The structure of the element is as follows:
        {
            'node_uri': `uuid.UUID`
                URI of the node
            'parent_uri': `Optional[uuid.UUID]`
                URI of the parent node
            'path_id': `List[uuid.UUID]`
                List of stroke ids
            'bounding_box': `Dict[str, float]`
                Bounding box
            'type': `str`
                Type
            'attributes': `List[Tuple[str, Any]]`
                List of attributes
        }

    Example
    -------
    >>> from uim.model.ink import InkModel
    >>> from uim.codec.parser.uim import UIMParser
    >>> from uim.model.helpers.schema_content_extractor import uim_schema_semantics_from
    >>> parser = UIMParser()
    >>> ink_model = parser.parse("path/to/ink.uim")
    >>> schema_semantics = uim_schema_semantics_from(ink_model)
    >>> print(schema_semantics)
    >>> [
    >>>   {
    >>>     'node_uri': UUID('16918b3f-b192-466e-83a3-54835ddfff11'),
    >>>     'parent_uri': UUID('16918b3f-b192-466e-83a3-54835ddfff11'),
    >>>     'path_id': [UUID('16918b3f-b192-466e-83a3-54835ddfff11')],
    >>>     'bounding_box': {'x': 175.71, 'y': 150.65, 'width': 15.91, 'height': 27.018},
    >>>     'type': 'will:math-structures/0.1/Symbol',
    >>>      'attributes': [('symbolType', 'Numerical'), ('representation', 'e')]
    >>>   }, ... ]
    """

    elements: List[Dict[str, Any]] = []
    entity_map: Dict[str, Dict[str, Any]] = {}
    root: InkNode = ink_model.view_root(str(semantic_view))

    # Iterate for triples with triple list and look for words
    for s in ink_model.knowledge_graph.statements:
        if s.predicate == schema.IS:
            if s.subject not in entity_map:
                entity_map[s.subject] = {
                    'type': s.object,
                    'attributes': [(st.predicate, st.object)
                                   for st in ink_model.knowledge_graph.all_statements_for(s.subject)
                                   if st.predicate != schema.IS]
                }
    ink_model.calculate_bounds_recursively(root)
    # Iterate
    for node in PreOrderEnumerator(root):
        if node.uri in entity_map:
            path_ids: List[uuid.UUID] = __collected_stroke_ids__(node)
            entry: Dict[str, Any] = {
                'path_id': path_ids, 'node_uri': node.uri, 'parent_uri': node.parent.uri if node.parent else None,
                'bounding_box': {
                    'x': node.group_bounding_box.x,
                    'y': node.group_bounding_box.y,
                    'width': node.group_bounding_box.width,
                    'height': node.group_bounding_box.height
                },
                'type': entity_map[node.uri]['type'],
                'attributes': entity_map[node.uri]['attributes']
            }
            elements.append(entry)
    return elements



# -*- coding: utf-8 -*-
# Copyright © 2024-present Wacom Authors. All Rights Reserved.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
from pathlib import Path

from uim.codec.parser.uim import UIMParser
from uim.model.helpers.serialize import json_encode
from uim.model.ink import InkModel

if __name__ == '__main__':
    parser: UIMParser = UIMParser()
    ink_model: InkModel = parser.parse(Path(__file__).parent / '..' / 'ink' / 'special' / 'ink.uim')
    # Convert the model to JSON
    with open('ink.json', 'w') as f:
        # json_encode is a helper function to convert the model to JSON
        f.write(json_encode(ink_model))




# -*- coding: utf-8 -*-
# Copyright © 2021-present Wacom Authors. All Rights Reserved.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
import string
import uuid
from typing import List, Tuple, Dict, Any, Optional

from uim.codec.parser.uim import UIMParser
from uim.model.helpers.schema_content_extractor import __collected_stroke_ids__
from uim.model.helpers.treeiterator import PreOrderEnumerator
from uim.model.ink import InkModel
from uim.model.semantics.node import InkNode, StrokeGroupNode, StrokeNode
from uim.model.semantics.schema import CommonViews, SegmentationSchema, NamedEntityRecognitionSchema

SEMANTIC_PREFIX: str = 'will://semantic/3.0/'
LEGACY_WORD: str = 'will://segmentation/3.0/Word'
LEGACY_TEXT_LINE: str = 'will://segmentation/3.0/TextLine'
LEGACY_HAS_CONTENT: str = f'{SEMANTIC_PREFIX}is'
LEGACY_HAS_ALT_CONTENT: str = f'{SEMANTIC_PREFIX}hasAlt'
LEGACY_HAS_NAMED_ENTITY: str = f'{SEMANTIC_PREFIX}hasNamedEntityDefinition'

LEGACY_NE_MAPPING: Dict[str, str] = {
    f'{SEMANTIC_PREFIX}nerBackend': 'provider',
    f'{SEMANTIC_PREFIX}hasUri': 'uri',
    f'{SEMANTIC_PREFIX}hasLabel': 'label',
    f'{SEMANTIC_PREFIX}hasThumb': 'image',
    f'{SEMANTIC_PREFIX}hasAbstract': 'description',
    f'{SEMANTIC_PREFIX}hasConfidence': 'confidence'
}

NE_MAPPING: Dict[str, str] = {
    NamedEntityRecognitionSchema.HAS_LANGUAGE.lower(): 'language',
    NamedEntityRecognitionSchema.HAS_PROVIDER.lower(): 'provider',
    NamedEntityRecognitionSchema.HAS_CREATION_DATE.lower(): 'creationDate',
    NamedEntityRecognitionSchema.HAS_ABSTRACT_TEXT.lower(): 'description',
    NamedEntityRecognitionSchema.HAS_THUMBNAIL_URL.lower(): 'image',
    NamedEntityRecognitionSchema.HAS_URI.lower(): 'uri',
    NamedEntityRecognitionSchema.HAS_LABEL.lower(): 'label',
    NamedEntityRecognitionSchema.HAS_ENTITY_TYPE.lower(): 'ontologyType'
}


def uim_extract_text_and_semantics(uim_bytes: bytes, hwr_view: str = CommonViews.HWR_VIEW.value) \
        -> tuple[list[dict], Any, str]:
    """
    Extracting the text from Universal Ink Model.

    Parameters
    ----------
    uim_bytes: `bytes`
        Byte array with RIFF file from Universal Ink Model
    hwr_view: `str`
       HWR view.

    Returns
    -------
    words: `List[dict]`
        List of words. Each word has its own dict containing the text, bounding box, and all alternatives.
    entities: `Dict[str, List[dict]]`
        Dictionary of entities. Each entity has its own dict containing the label, instance, and path ids.
    text: `str`
        Text extracted from the Universal Ink Model.

    Raises
    ------
        `InkModelException`
            If the Universal Ink Model does not contain the view with the requested view name.
    """
    uim_parser: UIMParser = UIMParser()
    ink_object: InkModel = uim_parser.parse(uim_bytes)
    return uim_extract_text_and_semantics_from(ink_object, hwr_view)


def uim_extract_text_and_semantics_from(ink_model: InkModel, hwr_view: str = CommonViews.HWR_VIEW.value)\
        -> Tuple[List[Dict[str, Any]], Dict[str, List[Dict[str, Any]]], str]:
    """
    Extracting the text from Universal Ink Model.

    Parameters
    ----------
    ink_model: InkModel
        Universal Ink Model
    hwr_view: str
       Name of the HWR view.

    Returns
    -------
    words: `List[dict]`
        List of words. Each word has its own dict containing the text, bounding box, and all alternatives.
    entities: `Dict[str, List[dict]]`
        Dictionary of entities. Each entity has its own dict containing the label, instance, and path ids.
    text: `str`
        Text extracted from the Universal Ink Model.

    Raises
    ------
        `InkModelException`
            If the Universal Ink Model does not contain the view with the requested view name.

     Examples
    --------
    >>> from pathlib import Path
    >>> from typing import Dict, Any
    >>> from uim.codec.parser.uim import UIMParser
    >>> from uim.model.helpers.text_extractor import uim_extract_text_and_semantics_from
    >>> path: Path = Path('ink_3_1_0.uim')
    >>> parser: UIMParser = UIMParser()
    >>> ink_model: InkModel = parser.parse(path)
    >>> words, entities, text = uim_extract_text_and_semantics_from(ink_model, CommonViews.HWR_VIEW.value)
    >>> for word in words:
    >>>     print(f"[text]: {word['text']}")
    >>>     print(f"[alternatives]: {'|'.join(word['alternatives'])}")
    >>>     print(f"[path ids]: {word['path_id']}")
    >>>     print(f"[word URI]: {word['word-uri']}")
    >>>     print(f"[bounding box]: x: {word['bounding_box']['x']}, y: {word['bounding_box']['y']}, "
    >>>           f"width: {word['bounding_box']['width']}, height: {word['bounding_box']['height']}")
    >>> for entity_uri, entity_hits in entities.items():
    >>>     print(f"[entity URI]: {entity_uri}")
    >>>     for entity in entity_hits:
    >>>         print(f"[entity]: {entity}")
    >>>         print(f"[path ids]: {entity['path_id']}")
    >>>         print(f"[instance]: {entity['instance']}")
    >>>         print(f"[provider]: {entity['provider']}")
    >>>         print(f"[uri]: {entity['uri']}")
    >>>         print(f"[image]: {entity['image']}")
    >>>         print(f"[description]: {entity['description']}")
    >>>         print(f"[label]: {entity['label']}")
    >>>         print(f"[bounding box]: x: {entity['bounding_box']['x']}, y: {entity['bounding_box']['y']}, "
    >>>               f"width: {entity['bounding_box']['width']}, height: {entity['bounding_box']['height']}")
    >>> print(f"[text]: {text}")
    """
    text: str = ''
    words: List[Dict[str, Any]] = []
    entity_map: Dict[str, Dict[str, Any]] = {}
    root: InkNode = ink_model.view_root(str(hwr_view))
    text_nodes: Dict[str, str] = {}
    text_alternatives: Dict[str, List[str]] = {}
    text_lines: List[str] = []
    entities: Dict[str, List[Dict[str, Any]]] = {}

    # Iterate for triples with triple list and look for words
    for s in ink_model.knowledge_graph.statements:
        if s.predicate.startswith(SegmentationSchema.HAS_ALT_CONTENT):
            if s.subject not in text_alternatives:
                text_alternatives[s.subject] = []
            text_alternatives[s.subject].append(s.object)
        if s.predicate == LEGACY_HAS_ALT_CONTENT:
            if s.subject not in text_alternatives:
                text_alternatives[s.subject] = []
            text_alternatives[s.subject].append(s.object)
        # Collect all words
        if s.object in [SegmentationSchema.WORD_OF_STROKES, SegmentationSchema.WORD]:
            all_statements = ink_model.knowledge_graph.all_statements_for(s.subject,
                                                                          predicate=SegmentationSchema.HAS_CONTENT)
            if len(all_statements) == 1:
                text_nodes[s.subject] = all_statements[0].object
        if s.object == LEGACY_WORD:
            all_statements = ink_model.knowledge_graph.all_statements_for(s.subject, predicate=LEGACY_HAS_CONTENT)
            if len(all_statements) == 1:
                text_nodes[s.subject] = all_statements[0].object
        # Collect all entities
        if s.predicate == NamedEntityRecognitionSchema.HAS_NAMED_ENTITY:
            all_statements = ink_model.knowledge_graph.all_statements_for(s.object)
            entity: Dict[str, Any] = {'instance': s.object}
            entity_map[s.subject] = entity
            for st in all_statements:
                if st.predicate.lower().startswith('haspart'):
                    entity_map[st.object] = entity
                elif st.predicate.lower() in NE_MAPPING:
                    entity[NE_MAPPING[st.predicate.lower()]] = st.object
        # Collect all entities for the legacy view
        if s.predicate == LEGACY_HAS_NAMED_ENTITY:
            all_statements = ink_model.knowledge_graph.all_statements_for(s.object)
            entity: Dict[str, Any] = {'instance': s.object}
            entity_map[s.subject] = entity
            for st in all_statements:
                if st.predicate in LEGACY_NE_MAPPING:
                    entity[LEGACY_NE_MAPPING[st.predicate]] = st.object
        # Collect all text lines
        if s.object == SegmentationSchema.TEXT_LINE:
            text_lines.append(s.subject)
        if s.object == LEGACY_TEXT_LINE:
            text_lines.append(s.subject)
    # Position
    pos: int = 0
    # Iterate
    for node in PreOrderEnumerator(root):
        if node.uri in text_lines:
            for word_node in node.children:
                path_ids: List[uuid.UUID] = __collected_stroke_ids__(node)
                if word_node.uri in text_nodes:
                    alternatives: List[str] = text_alternatives.get(word_node.uri, [])
                    t = text_nodes[word_node.uri]
                    if t in string.punctuation or pos == 0:
                        text += t
                    else:
                        text += f' {t}'
                    words.append({
                        'alternatives': alternatives, 'text': t, 'path_id': path_ids, 'word-uri': word_node.uri,
                        'bounding_box': {
                            'x': word_node.group_bounding_box.x,
                            'y': word_node.group_bounding_box.y,
                            'width': word_node.group_bounding_box.width,
                            'height': word_node.group_bounding_box.height
                        }
                    })
                    # Position
                    pos += 1
                if word_node.uri in entity_map:
                    uri: str = entity_map[word_node.uri].get('uri', '')
                    if uri not in entities:
                        entities[uri] = []
                    entry: Dict[str, Any] = {
                        'path_id': path_ids,
                        'instance': entity_map[word_node.uri]['instance'],
                        'bounding_box': {
                            'x': word_node.group_bounding_box.x,
                            'y': word_node.group_bounding_box.y,
                            'width': word_node.group_bounding_box.width,
                            'height': word_node.group_bounding_box.height
                        }
                    }
                    for key in entity_map[word_node.uri]:
                        entry[key] = entity_map[word_node.uri][key]
                    entities[uri].append(entry)
            text += '\n'
    ne_view: Optional[InkNode] = None
    if ink_model.has_tree(CommonViews.LEGACY_NER_VIEW.value):
        # In the legacy view, the named entities are in a separate view
        ne_view = ink_model.view_root(CommonViews.LEGACY_NER_VIEW.value)
    if ink_model.has_tree(CommonViews.NER_VIEW.value):
        # In the new view, the named entities are in a separate view
        ne_view = ink_model.view_root(CommonViews.NER_VIEW.value)
    if ne_view is not None:
        for node in PreOrderEnumerator(ne_view):
            if node.uri in entity_map:
                path_ids: List[uuid.UUID] = __collected_stroke_ids__(node)
                uri: str = entity_map[node.uri]['uri'] if 'uri' in entity_map[node.uri] else ''
                if uri not in entities:
                    entities[uri] = []
                entry: Dict[str, Any] = {
                    'path_id': path_ids,
                    'instance': entity_map[node.uri]['instance'],
                    'bounding_box': {
                        'x': node.group_bounding_box.x,
                        'y': node.group_bounding_box.y,
                        'width': node.group_bounding_box.width,
                        'height': node.group_bounding_box.height
                    }
                }
                for key in entity_map[node.uri]:
                    entry[key] = entity_map[node.uri][key]
                entities[uri].append(entry)
    if text.endswith('\n'):
        text = text[:-1]
    return words, entities, text



# -*- coding: utf-8 -*-
# Copyright © 2023-present Wacom Authors. All Rights Reserved.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
from abc import ABC, abstractmethod
from typing import List, Dict, Any, Optional

import numpy as np

from uim.model.base import InkModelException
from uim.model.helpers.policy import HandleMissingDataPolicy
from uim.model.ink import InkModel, InkTree, logger
from uim.model.inkdata.strokes import Stroke
from uim.model.inkinput.inputdata import InkSensorType, InputContext, SensorContext, SensorChannel
from uim.model.inkinput.sensordata import SensorData, ChannelData


def safe_zero_div(x: float, y: float) -> float:
    """
    Safely divide two numbers. If the denominator is zero, return zero.
    Parameters
    ----------
    x: float
        Numerator
    y: float
        Denominator

    Returns
    -------
    division: float
        x / y or 0. if y == 0.
    """
    try:
        return x / y
    except ZeroDivisionError:
        return 0.


class ModelAnalyzer(ABC):
    """
    Model analyzer
    ==============

    Abstract class for model analysis.
    """
    KNOWN_TYPE_PREDICATES: List[str] = ["@", "http://www.w3.org/1999/02/22-rdf-syntax-ns#type"]
    """Known type predicates"""

    @staticmethod
    @abstractmethod
    def analyze(model: InkModel) -> Dict[str, Any]:
        """
        Analyze the model.
        Parameters
        ----------
        model: InkModel
            Ink model to analyze

        Returns
        -------
        summary: Dict[str, Any]
            Summary of the analysis
        """

    @staticmethod
    def __assume_view_type_predicate__(model: InkModel, view: InkTree) -> Optional[str]:
        statements = model.knowledge_graph.all_statements_for(subject=view.root.uri)
        for statement in statements:
            if statement.predicate in ModelAnalyzer.KNOWN_TYPE_PREDICATES:
                return statement.predicate
        return None

    @staticmethod
    def __extract_sensor_data_info__(model: InkModel, stroke: Stroke, stats: Dict[str, Any]):
        try:
            sd: SensorData = model.sensor_data.sensor_data_by_id(stroke.sensor_data_id)
            ic: InputContext = model.input_configuration.get_input_context(sd.input_context_id)
            sc: SensorContext = model.input_configuration.get_sensor_context(ic.sensor_context_id)
        except InkModelException as e:
            logger.error(f"Error while extracting sensor data info: {e}")
            return

        for scc in sc.sensor_channels_contexts:
            for channel in scc.channels:
                channel: SensorChannel = channel
                channel_type = channel.type
                existing_channel = stats['sensor_channels'].get(channel_type.name)

                if existing_channel is None:
                    stats['sensor_channels'][channel_type.name] = {
                        'strokes_count': 0, 'percent': 0, 'values': [], "metric": channel.metric.name,
                        "resolution": channel.resolution, "precision": channel.precision,
                        "channel_min": channel.min, "channel_max": channel.max
                    }

                stats['sensor_channels'][channel_type.name]['strokes_count'] += 1

                values = sd.get_data_by_id(channel.id).values
                if channel.type == InkSensorType.TIMESTAMP:
                    values = [v + sd.timestamp for v in values]

                stats['sensor_channels'][channel_type.name]['values'].extend(values)

    @staticmethod
    def __post_process_sensor_channels_info__(stats):
        for k, v in stats['sensor_channels'].items():
            stats['sensor_channels'][k]["min"] = min(v['values'])
            stats['sensor_channels'][k]["max"] = max(v['values'])
            stats['sensor_channels'][k]["mean"] = np.mean(v['values'])
            stats['sensor_channels'][k]["median"] = np.median(v['values'])
            stats['sensor_channels'][k].pop('values', None)


def get_channel_data_values(ink_model: InkModel, stroke: Stroke, ink_sensor_type: InkSensorType) -> List[float]:
    """
    Get channel data values for a given stroke and sensor type.
    Parameters
    ----------
    ink_model: InkModel
        Ink model
    stroke: Stroke
        Stroke
    ink_sensor_type: InkSensorType
        Sensor type

    Returns
    -------
    channel_data: List[float]
        Channel data values
    """
    channel_data: Optional[ChannelData] = get_channel_data_instance(ink_model, stroke, ink_sensor_type)
    if channel_data is None:
        return []

    if ink_sensor_type == InkSensorType.TIMESTAMP:
        sd: SensorData = ink_model.sensor_data.sensor_data_by_id(stroke.sensor_data_id)
        return [v + sd.timestamp for v in channel_data.values]
    return channel_data.values.copy()


def get_channel_data_instance(ink_model: InkModel, stroke: Stroke, ink_sensor_type: InkSensorType) \
        -> Optional[ChannelData]:
    """
    Get channel data instance for a given stroke and sensor type.
    Parameters
    ----------
    ink_model: InkModel
        Ink model
    stroke: Stroke
        Stroke
    ink_sensor_type: InkSensorType
        Sensor type

    Returns
    -------
    channel_data: Optional[ChannelData]
        Channel data instance
    """
    sd: SensorData = ink_model.sensor_data.sensor_data_by_id(stroke.sensor_data_id)
    sc: Optional[SensorChannel] = None
    input_context: InputContext = ink_model.input_configuration.get_input_context(sd.input_context_id)
    if input_context is not None:
        sensor_context = ink_model.input_configuration.get_sensor_context(input_context.sensor_context_id)
        if sensor_context is not None:

            if sensor_context.has_channel_type(ink_sensor_type):
                sc = sensor_context.get_channel_by_type(ink_sensor_type)

    if sd is None or sc is None or sd.get_data_by_id(sc.id) is None:
        return None
    return sd.get_data_by_id(sc.id)


def as_strided_array(ink_model: InkModel, stroke: Stroke, handle_missing_data=HandleMissingDataPolicy.FILL_WITH_ZEROS) \
        -> Optional[List[float]]:
    """
    Convert stroke to strided array.

    Parameters
    ----------
    ink_model: InkModel
        Ink model
    stroke: Stroke
        Stroke
    handle_missing_data: HandleMissingDataPolicy
        Handle missing data policy

    Returns
    -------
    points: Optional[List[float]]
        Strided array
    """
    # Remove the first and last element, which are added by the spline producer
    xs: List[float] = stroke.splines_x[1:-1]
    ys: List[float] = stroke.splines_y[1:-1]

    if stroke.sensor_data_id is None:
        ts: List[float] = []
        ps: List[float] = []
    else:
        ts: List[float] = get_channel_data_values(ink_model, stroke, InkSensorType.TIMESTAMP)
        ps: List[float] = get_channel_data_values(ink_model, stroke, InkSensorType.PRESSURE)

    # Handle missing timestamp according to policy
    if len(ts) == 0:
        if handle_missing_data == HandleMissingDataPolicy.FILL_WITH_ZEROS:
            ts = [0 for i in range(len(xs))]
        elif handle_missing_data == HandleMissingDataPolicy.FILL_WITH_NAN:
            NaN = float("nan")
            ts = [NaN for i in range(len(xs))]
        elif handle_missing_data == HandleMissingDataPolicy.SKIP_STROKE:
            return None
        elif handle_missing_data == HandleMissingDataPolicy.THROW_EXCEPTION:
            raise ValueError("There is no timestamp data for this stroke.")

    target_len: int = len(ts) if len(ts) > 0 else len(xs)

    # Handle missing pressure according to policy
    if len(ps) == 0:
        if handle_missing_data == HandleMissingDataPolicy.FILL_WITH_ZEROS:
            ps = [0 for i in range(target_len)]
        elif handle_missing_data == HandleMissingDataPolicy.FILL_WITH_NAN:
            NaN: float = float("nan")
            ps = [NaN for i in range(target_len)]
        elif handle_missing_data == HandleMissingDataPolicy.SKIP_STROKE:
            return None
        elif handle_missing_data == HandleMissingDataPolicy.THROW_EXCEPTION:
            raise ValueError("There is no pressure data for this stroke.")

    xs = xs[0:target_len]
    ys = ys[0:target_len]

    points: List[float] = []

    sensor_data_mapping = stroke.sensor_data_mapping

    if len(sensor_data_mapping) == 0:  # Mapping is 1:1
        limit: int = min(stroke.sensor_data_offset + len(xs), len(ts))
        sensor_data_mapping = range(stroke.sensor_data_offset, limit)

    i: int = 0

    for map_i in sensor_data_mapping:
        points.append(xs[i])
        points.append(ys[i])

        if len(ts) == 0:
            points.append(0)
        else:
            points.append(ts[map_i])

        if len(ps) == 0:
            points.append(0)
        else:
            points.append(ps[map_i])

        i += 1
    return points
